#!/usr/bin/env python
# -*- coding=UTF-8 -*-
# vim: fileencoding=UTF-8 tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -----------------------------------------------------------------------------
# Document Description
"""**Test Multiton**"""

# -----------------------------------------------------------------------------
# Module Level Dunders
__title__ = "Test Multiton"
__module__ = "test_multiton.py"
__author__ = "Martin Abel"
__maintainer__ = "Martin Abel"
__credits__ = ["Martin Abel"]
__company__ = "eVation"
__email__ = "python@evation.eu"
__url__ = "http://evation.eu"
__copyright__ = f"(c) copyright 2020-2023, {__company__}"
__created__ = "2020-01-01"
__version__ = "0.9.0.1"
__date__ = "2023-06-21"
__time__ = "13:37:48"
__state__ = "Beta"
__license__ = "PSF"

# -----------------------------------------------------------------------------
# Libraries & Modules
import unittest

from decoratory.multiton import Multiton


# -----------------------------------------------------------------------------
# Test Class
# noinspection PyPep8Naming
class TestMultiton(unittest.TestCase):
    class TestClass:
        """A simple test class with some parameters"""

        def __init__(self, name: str, foo=None, bar="any stuff"):
            self.name = name
            self.foo = foo
            self.bar = bar

        def __repr__(self):
            return f"{self.__class__.__name__}(" \
                   f"'{self.name}', {self.foo}, '{self.bar}')"

    class AltClass:
        """An alternative test class with some parameters"""

        def __init__(self, name: str, foo=None, bar="other stuff"):
            self.name = name
            self.foo = foo
            self.bar = bar

        def __repr__(self):
            return f"{self.__class__.__name__}(" \
                   f"'{self.name}', {self.foo}, '{self.bar}')"

    def setUp(self):
        """Setup"""
        pass

    def tearDown(self):
        """Shutdown"""
        pass

    def test_decoration_without_brackets(self):
        """Unittest: Decoration without brackets
            --> No key: Multiton == Singleton
        """

        TestClass = TestMultiton.TestClass
        TestClassMultiton = Multiton(TestClass)

        # Instances
        a = TestClassMultiton(name="a")  # New single object a
        b = TestClassMultiton(name="b")  # Multiton: b is a
        c = TestClassMultiton(name="b")  # Multiton: c is b is a

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIsInstance(c, TestClass)
        self.assertIs(a, b)
        self.assertIs(b, c)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('a', None, 'any stuff')")
        self.assertEqual(a, b)
        self.assertEqual(b, c)
        self.assertEqual(TestClass.__doc__, TestClassMultiton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)
        self.assertEqual(TestClass.__doc__, c.__doc__)

    def test_decoration_without_brackets_default_args(self):
        """Unittest: Decoration without brackets plus default args
            --> No key: Multiton == Singleton
        """
        TestClass = TestMultiton.TestClass
        TestClassMultiton = Multiton(TestClass, "a")

        # Instances
        a = TestClassMultiton()  # New single default object a
        b = TestClassMultiton(name="b")  # Multiton: b is a
        c = TestClassMultiton(name="b")  # Multiton: c is b is a

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIsInstance(c, TestClass)
        self.assertIs(a, b)
        self.assertIs(b, c)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('a', None, 'any stuff')")
        self.assertEqual(a, b)
        self.assertEqual(b, c)
        self.assertEqual(TestClass.__doc__, TestClassMultiton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)
        self.assertEqual(TestClass.__doc__, c.__doc__)

    def test_decoration_without_brackets_default_kwargs(self):
        """Unittest: Decoration without brackets plus default kwargs
            --> No key: Multiton == Singleton
        """
        TestClass = TestMultiton.TestClass
        TestClassMultiton = Multiton(TestClass, name="a")

        # Instances
        a = TestClassMultiton()  # New single object a
        b = TestClassMultiton(name="b")  # Multiton: b is a
        c = TestClassMultiton(name="b")  # Multiton: c is b is a

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIsInstance(c, TestClass)
        self.assertIs(a, b)
        self.assertIs(b, c)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('a', None, 'any stuff')")
        self.assertEqual(a, b)
        self.assertEqual(b, c)
        self.assertEqual(TestClass.__doc__, TestClassMultiton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)
        self.assertEqual(TestClass.__doc__, c.__doc__)

    def test_decoration_with_empty_brackets(self):
        """Unittest: Decoration with empty brackets
            --> No key: Multiton == Singleton
        """
        TestClass = TestMultiton.TestClass
        TestClassMultiton = Multiton()(TestClass)

        # Instances
        a = TestClassMultiton(name="a")  # New single object a
        b = TestClassMultiton(name="b")  # Multiton: b is a
        c = TestClassMultiton(name="b")  # Multiton: c is b is a

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIsInstance(c, TestClass)
        self.assertIs(a, b)
        self.assertIs(b, c)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('a', None, 'any stuff')")
        self.assertEqual(a, b)
        self.assertEqual(b, c)
        self.assertEqual(TestClass.__doc__, TestClassMultiton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)
        self.assertEqual(TestClass.__doc__, c.__doc__)

    def test_decoration_with_invalid_parameter(self):
        """Unittest: Decoration with an invalid key object
            --> key = None: Multiton == Singleton
        """
        TestClass = TestMultiton.TestClass
        TestClassMultiton = Multiton(key=list())(TestClass)  # mutable key!

        # Instances
        a = TestClassMultiton(name="a")  # New single object a
        b = TestClassMultiton(name="b")  # Multiton: b is a
        c = TestClassMultiton(name="b")  # Multiton: c is b is a

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIsInstance(c, TestClass)
        self.assertIs(a, b)
        self.assertIs(b, c)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('a', None, 'any stuff')")
        self.assertEqual(a, b)
        self.assertEqual(b, c)
        self.assertEqual(TestClass.__doc__, TestClassMultiton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)
        self.assertEqual(TestClass.__doc__, c.__doc__)

    def test_decoration_with_object(self):
        """Unittest: Decoration with an object
            --> key = fixed: Multiton == Singleton
        """
        TestClass = TestMultiton.TestClass
        TestClassMultiton = Multiton(key=tuple())(TestClass)  # immutable key!

        # Instances
        a = TestClassMultiton(name="a")  # New single object a
        b = TestClassMultiton(name="b")  # Multiton: b is a
        c = TestClassMultiton(name="b")  # Multiton: c is b is a

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIsInstance(c, TestClass)
        self.assertIs(a, b)
        self.assertIs(b, c)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('a', None, 'any stuff')")
        self.assertEqual(a, b)
        self.assertEqual(b, c)
        self.assertEqual(TestClass.__doc__, TestClassMultiton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)
        self.assertEqual(TestClass.__doc__, c.__doc__)

    def test_decoration_with_valid_parameter_id(self):
        """Unittest: Decoration with a valid key parameter: args[0]
            -> key = "name" attribute from args[0]
        """
        TestClass = TestMultiton.TestClass
        TestClassMultiton = Multiton(key="{0}".format)(TestClass)  # args[0]

        # Instances
        a = TestClassMultiton("a")  # New single object a
        b = TestClassMultiton("b")  # New single object b
        c = TestClassMultiton("b")  # Multiton: c is b

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIsInstance(c, TestClass)
        self.assertIsNot(a, b)
        self.assertIsNot(a, c)
        self.assertIs(b, c)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('b', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('b', None, 'any stuff')")
        self.assertNotEqual(a, b)
        self.assertNotEqual(a, c)
        self.assertEqual(b, c)
        self.assertEqual(TestClass.__doc__, TestClassMultiton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)
        self.assertEqual(TestClass.__doc__, c.__doc__)

    def test_decoration_with_valid_parameter_name(self):
        """Unittest: Decoration with a valid key parameter: kwargs['name']
            --> key = "name" attribute from kwargs['name']
        """
        TestClass = TestMultiton.TestClass
        TestClassMultiton = Multiton(key="{name}".format)(
            TestClass)  # kwargs['name']

        # Instances
        a = TestClassMultiton(name="a")  # New single object a
        b = TestClassMultiton(name="b")  # New single object b
        c = TestClassMultiton(name="b")  # Multiton: c is b

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIsInstance(c, TestClass)
        self.assertIsNot(a, b)
        self.assertIsNot(a, c)
        self.assertIs(b, c)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('b', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('b', None, 'any stuff')")
        self.assertNotEqual(a, b)
        self.assertNotEqual(a, c)
        self.assertEqual(b, c)
        self.assertEqual(TestClass.__doc__, TestClassMultiton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)
        self.assertEqual(TestClass.__doc__, c.__doc__)

    def test_decoration_with_valid_parameter_inherited(self):
        """Unittest: Decoration with a valid key parameter: lambda name: name
            --> key = "name" attribute from lambda name: name
        """

        class TestClass(TestMultiton.TestClass):
            """Inherited test class"""
            pass

        TestClassMultiton = Multiton(key=lambda name: name)(TestClass)

        # Instances
        a = TestClassMultiton(name="a")  # New single object a
        b = TestClassMultiton(name="b")  # New single object b
        c = TestClassMultiton(name="b")  # Multiton: c is b

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIsInstance(c, TestClass)
        self.assertIsNot(a, b)
        self.assertIsNot(a, c)
        self.assertIs(b, c)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('b', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('b', None, 'any stuff')")
        self.assertNotEqual(a, b)
        self.assertNotEqual(a, c)
        self.assertEqual(b, c)
        self.assertEqual(TestClass.__doc__, TestClassMultiton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)
        self.assertEqual(TestClass.__doc__, c.__doc__)

    def test_decoration_with_callable_parameter(self):
        """Unittest: Decoration with a valid key callable: f = lambda name: name
            --> key = "name" attribute from function f = lambda name: name
        """
        f = lambda name: name  # Mapping as a function
        TestClass = TestMultiton.TestClass
        TestClassMultiton = Multiton(key=f)(TestClass)

        # Instances
        a = TestClassMultiton(name="a")  # New single object a
        b = TestClassMultiton(name="b")  # New single object b
        c = TestClassMultiton(name="b")  # Multiton: c is b

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIsInstance(c, TestClass)
        self.assertIsNot(a, b)
        self.assertIsNot(a, c)
        self.assertIs(b, c)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('b', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('b', None, 'any stuff')")
        self.assertNotEqual(a, b)
        self.assertNotEqual(a, c)
        self.assertEqual(b, c)
        self.assertEqual(TestClass.__doc__, TestClassMultiton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)
        self.assertEqual(TestClass.__doc__, c.__doc__)

    def test_decoration_with_callable_classmethod(self):
        """Unittest: Decoration with a valid key callable (class method)
            --> key = "name" attribute from @classmethod key_function(cls, name)
        """

        class TestClass(TestMultiton.TestClass):
            """Test class"""

            @classmethod
            def key_function(cls, name):
                """Key callable"""
                return name

        TestClassMultiton = Multiton(key=TestClass.key_function)(TestClass)

        # Instances
        a = TestClassMultiton(name="a")  # New single object a
        b = TestClassMultiton(name="b")  # New single object b
        c = TestClassMultiton(name="b")  # Multiton: c is b

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIsInstance(c, TestClass)
        self.assertIsNot(a, b)
        self.assertIsNot(a, c)
        self.assertIs(b, c)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('b', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('b', None, 'any stuff')")
        self.assertNotEqual(a, b)
        self.assertNotEqual(a, c)
        self.assertEqual(b, c)
        self.assertEqual(TestClass.__doc__, TestClassMultiton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)
        self.assertEqual(TestClass.__doc__, c.__doc__)

    def test_sealed(self):
        """Unittest: Decoration with a valid key parameter: "name" attribute
            --> Method test for seal(), unseal() and issealed()
        """
        TestClass = TestMultiton.TestClass
        TestClassMultiton = Multiton(key=lambda name: name)(TestClass)

        # Instances
        a = TestClassMultiton(name="a")  # New single object a
        b = TestClassMultiton(name="b")  # New single object b
        c = TestClassMultiton(name="b")  # Multiton: c is b

        # Tests
        self.assertFalse(TestClassMultiton.issealed())
        TestClassMultiton.seal()
        self.assertTrue(TestClassMultiton.issealed())
        self.assertRaises(KeyError, TestClassMultiton, name="x")
        self.assertTrue(TestClassMultiton.issealed())
        TestClassMultiton.unseal()
        self.assertFalse(TestClassMultiton.issealed())
        d = TestClassMultiton(name="d")  # New single object d

        self.assertFalse(TestClassMultiton.issealed())
        TestClassMultiton.seal()  # Accessible because of sealed=False
        self.assertTrue(TestClassMultiton.issealed())
        self.assertRaises(KeyError, TestClassMultiton, name="y")
        self.assertTrue(TestClassMultiton.issealed())
        TestClassMultiton.unseal()  # Accessible via every multiton instance
        self.assertFalse(TestClassMultiton.issealed())
        e = TestClassMultiton(name="e")  # New single object e

        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIsInstance(c, TestClass)
        self.assertIsInstance(d, TestClass)
        self.assertIsInstance(e, TestClass)
        self.assertIsNot(a, b)
        self.assertIsNot(a, c)
        self.assertIsNot(a, d)
        self.assertIsNot(a, e)
        self.assertIs(b, c)
        self.assertIsNot(b, d)
        self.assertIsNot(c, d)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('b', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('b', None, 'any stuff')")
        self.assertEqual(repr(d), "TestClass('d', None, 'any stuff')")
        self.assertEqual(repr(e), "TestClass('e', None, 'any stuff')")
        self.assertNotEqual(a, b)
        self.assertNotEqual(a, c)
        self.assertNotEqual(a, d)
        self.assertNotEqual(a, e)
        self.assertEqual(b, c)
        self.assertNotEqual(b, d)
        self.assertNotEqual(c, d)
        self.assertEqual(TestClass.__doc__, TestClassMultiton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)
        self.assertEqual(TestClass.__doc__, c.__doc__)
        self.assertEqual(TestClass.__doc__, d.__doc__)
        self.assertEqual(TestClass.__doc__, e.__doc__)

    def test_reset_instances(self):
        """Unittest: Decoration with a valid key parameter: "name" attribute
            --> Method test for instances and reset()
        """
        TestClass = TestMultiton.TestClass
        TestClassMultiton = Multiton(key=lambda name: name,
                                     resettable=True)(TestClass)

        # Instances
        a = TestClassMultiton(name="a")  # New single object a
        b = TestClassMultiton(name="b")  # New single object b

        # Tests
        self.assertDictEqual(TestClassMultiton.instances, {'a': a, 'b': b})
        TestClassMultiton.reset()
        self.assertDictEqual(TestClassMultiton.instances, {})
        self.assertFalse(TestClassMultiton.issealed())
        a = TestClassMultiton(name="a")  # New single object a
        self.assertDictEqual(TestClassMultiton.instances, {'a': a})
        TestClassMultiton.seal()
        self.assertTrue(TestClassMultiton.issealed())
        try:
            TestClassMultiton(name="b")
        except KeyError as ex:
            self.assertTupleEqual(ex.args, ('TestClass is sealed.', 'b'))
        self.assertDictEqual(TestClassMultiton.instances, {'a': a})
        TestClassMultiton.unseal()
        self.assertFalse(TestClassMultiton.issealed())
        c = TestClassMultiton(name="c")
        self.assertDictEqual(TestClassMultiton.instances, {'a': a, 'c': c})
        TestClassMultiton.instances.pop("a")
        self.assertDictEqual(TestClassMultiton.instances, {'c': c})
        TestClassMultiton.reset()
        self.assertDictEqual(TestClassMultiton.instances, {})

        TestClassMultiton.instances = {'a': a}
        self.assertDictEqual(TestClassMultiton.instances, {'a': a})
        a = TestClassMultiton(name="a")
        self.assertDictEqual(TestClassMultiton.instances, {'a': a})
        TestClassMultiton(name="a")
        self.assertDictEqual(TestClassMultiton.instances, {'a': a})

    def test_concurrency(self):
        """Unittest: Decoration with a valid key parameter: "name" attribute
            --> Two parallel multitons do not interfere with each other!
        """
        TestClass = TestMultiton.TestClass
        TestClassMultiton = Multiton(key=lambda name: name)(TestClass)
        AltClass = TestMultiton.AltClass
        AltClassMultiton = Multiton(key=lambda name: name)(AltClass)

        # Instances
        a = TestClassMultiton(name="a")  # New single object a
        b = TestClassMultiton(name="b")  # New single object b
        c = TestClassMultiton(name="b")  # Multiton: c is b
        x = AltClassMultiton(name="x")  # New single object x
        y = AltClassMultiton(name="y")  # New single object y
        z = AltClassMultiton(name="y")  # Multiton: z is y

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIsInstance(c, TestClass)
        self.assertIsNot(a, b)
        self.assertIsNot(a, c)
        self.assertIs(b, c)
        self.assertIsInstance(x, AltClass)
        self.assertIsInstance(y, AltClass)
        self.assertIsInstance(z, AltClass)
        self.assertIsNot(x, y)
        self.assertIsNot(x, z)
        self.assertIs(y, z)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('b', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('b', None, 'any stuff')")
        self.assertEqual(repr(x), "AltClass('x', None, 'other stuff')")
        self.assertEqual(repr(y), "AltClass('y', None, 'other stuff')")
        self.assertEqual(repr(z), "AltClass('y', None, 'other stuff')")
        self.assertNotEqual(a, b)
        self.assertNotEqual(a, c)
        self.assertEqual(b, c)
        self.assertNotEqual(x, y)
        self.assertNotEqual(x, z)
        self.assertEqual(y, z)

        self.assertEqual(TestClass.__doc__, TestClassMultiton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)
        self.assertEqual(TestClass.__doc__, c.__doc__)
        self.assertEqual(AltClass.__doc__, AltClassMultiton.__doc__)
        self.assertEqual(AltClass.__doc__, x.__doc__)
        self.assertEqual(AltClass.__doc__, y.__doc__)
        self.assertEqual(AltClass.__doc__, z.__doc__)


# -----------------------------------------------------------------------------
# Execution
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding=UTF-8 -*-
# vim: fileencoding=UTF-8 tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -----------------------------------------------------------------------------
# Document Description
"""**Test Observer**"""

# -----------------------------------------------------------------------------
# Module Level Dunders
__title__ = "Test Observer"
__module__ = "test_observer.py"
__author__ = "Martin Abel"
__maintainer__ = "Martin Abel"
__credits__ = ["Martin Abel"]
__company__ = "eVation"
__email__ = "python@evation.eu"
__url__ = "http://evation.eu"
__copyright__ = f"(c) copyright 2020-2023, {__company__}"
__created__ = "2020-01-01"
__version__ = "0.9.0.2"
__date__ = "2023-06-23"
__time__ = "08:53:07"
__state__ = "Beta"
__license__ = "PSF"

# -----------------------------------------------------------------------------
# Libraries & Modules
import unittest

from decoratory.observer import (BaseObservable, BaseObserver,
                                 Observable, Observer)
from decoratory.basic import Activation, F, X


# -----------------------------------------------------------------------------
# Test functions
def cmp(iter1, iter2):
    """Compare *all* elements of F with a tuple"""
    if len(iter1) != len(iter2):
        return False
    else:
        iter1 = list(iter1)
        iter2 = list(iter2)
        iter1.sort()
        iter2.sort()
        for left, right in zip(iter1, iter2):
            if left.callee != right.callee or \
                    left.callee_args != right.callee_args or \
                    left.callee_kwargs != right.callee_kwargs:
                return False
        return True


# -----------------------------------------------------------------------------
# Test Class
# noinspection PyPep8Naming
# noinspection PyUnresolvedReferences
# noinspection PyArgumentEqualDefault
# noinspection PyPropertyAccess
class TestObserver(unittest.TestCase):

    def setUp(self):
        """Preparation"""
        pass

    def tearDown(self):
        """Wrap-up"""
        pass

    def test_nodecoration_noparams_without_brackets(self):
        """Unittest: Functions - no decoration, no parameters, no brackets"""

        # Result list
        res = list()

        # Observable
        def dog(act: str = "Brrr!") -> None:
            """A dog function"""
            res.append(f"{dog.__name__} acts '{act}'")

        # Observer
        def person(say: str = "Hello?") -> None:
            """A person function"""
            res.append(f"{person.__name__} says '{say}'")

        # ---------------------------------------------------------------------
        # Test setup: Observable, only
        dog_obsl = Observable(dog)

        # Checks for dog_obsl
        self.assertTrue(isinstance(dog_obsl, Observable))
        self.assertIs(dog_obsl.BaseClass, BaseObservable)
        self.assertTrue(isinstance(dog_obsl.observable, dog_obsl.BaseClass))
        self.assertTrue(hasattr(dog_obsl.observable, 'register'))
        self.assertTrue(hasattr(dog_obsl.observable, 'unregister'))
        self.assertTrue(hasattr(dog_obsl.observable, 'dispatch'))
        self.assertTrue(hasattr(dog_obsl.observable, 'observers'))
        self.assertTupleEqual(dog_obsl.observable.args, ())
        self.assertDictEqual(dog_obsl.observable.kwargs, {})
        self.assertIs(dog_obsl.activate, Activation.AFTER)
        self.assertListEqual(dog_obsl.methods, [])
        self.assertIs(dog_obsl.substitute.callee, dog)
        self.assertTupleEqual(dog_obsl.substitute.callee_args, ())
        self.assertDictEqual(dog_obsl.substitute.callee_kwargs, {})
        self.assertIs(dog_obsl.__annotations__, dog.__annotations__)
        self.assertIs(dog_obsl.__doc__, dog.__doc__)
        self.assertIs(dog_obsl.__name__, dog.__name__)

        # Calls
        res.clear()
        person()
        self.assertListEqual(res, ["person says 'Hello?'"])
        res.clear()
        dog_obsl()
        self.assertListEqual(res, ["dog acts 'Brrr!'"])

        # ---------------------------------------------------------------------
        # Test setup: Observable & Observer
        prs_obsr = Observer(person)

        # Checks for prs_obsr
        self.assertTrue(isinstance(prs_obsr, Observer))
        self.assertIs(prs_obsr.BaseClass, BaseObserver)
        self.assertTrue(isinstance(prs_obsr.observer, prs_obsr.BaseClass))
        self.assertTupleEqual(prs_obsr.observer.args, ())
        self.assertDictEqual(prs_obsr.observer.kwargs, {})
        self.assertListEqual(prs_obsr.methods, [])
        self.assertListEqual(prs_obsr.observables, [])
        self.assertIs(prs_obsr.substitute.callee, person)
        self.assertTupleEqual(prs_obsr.substitute.callee_args, ())
        self.assertDictEqual(prs_obsr.substitute.callee_kwargs, {})
        self.assertIs(prs_obsr.__annotations__, person.__annotations__)
        self.assertIs(prs_obsr.__doc__, person.__doc__)
        self.assertIs(prs_obsr.__name__, person.__name__)

        # ---------------------------------------------------------------------
        # Calls
        res.clear()
        prs_obsr()
        self.assertListEqual(res, ["person says 'Hello?'"])
        res.clear()
        dog_obsl()
        self.assertListEqual(res, ["dog acts 'Brrr!'"])

    def test_nodecoration_noparams_with_empty_brackets(self):
        """Unittest: Functions - no decoration, no parameters, empty brackets"""

        # Result list
        res = list()

        # Observable
        def dog(act: str = "Brrr!") -> None:
            """A dog function"""
            res.append(f"{dog.__name__} acts '{act}'")

        # Observer
        def person(say: str = "Hello?") -> None:
            """A person function"""
            res.append(f"{person.__name__} says '{say}'")

        # ---------------------------------------------------------------------
        # Test setup: Observable, only
        dog_obsl = Observable()(dog)

        # Checks for dog_obsl
        self.assertTrue(isinstance(dog_obsl, Observable))
        self.assertIs(dog_obsl.BaseClass, BaseObservable)
        self.assertTrue(isinstance(dog_obsl.observable, dog_obsl.BaseClass))
        self.assertTrue(hasattr(dog_obsl.observable, 'register'))
        self.assertTrue(hasattr(dog_obsl.observable, 'unregister'))
        self.assertTrue(hasattr(dog_obsl.observable, 'dispatch'))
        self.assertTrue(hasattr(dog_obsl.observable, 'observers'))
        self.assertTupleEqual(dog_obsl.observable.args, ())
        self.assertDictEqual(dog_obsl.observable.kwargs, {})
        self.assertIs(dog_obsl.activate, Activation.AFTER)
        self.assertListEqual(dog_obsl.methods, [])
        self.assertIs(dog_obsl.substitute.callee, dog)
        self.assertTupleEqual(dog_obsl.substitute.callee_args, ())
        self.assertDictEqual(dog_obsl.substitute.callee_kwargs, {})
        self.assertIs(dog_obsl.__annotations__, dog.__annotations__)
        self.assertIs(dog_obsl.__doc__, dog.__doc__)
        self.assertIs(dog_obsl.__name__, dog.__name__)

        # Calls
        res.clear()
        person()
        self.assertListEqual(res, ["person says 'Hello?'"])
        res.clear()
        dog_obsl()
        self.assertListEqual(res, ["dog acts 'Brrr!'"])

        # ---------------------------------------------------------------------
        # Test setup: Observable & Observer
        prs_obsr = Observer()(person)

        # Checks for prs_obsr
        self.assertTrue(isinstance(prs_obsr, Observer))
        self.assertIs(prs_obsr.BaseClass, BaseObserver)
        self.assertTrue(isinstance(prs_obsr.observer, prs_obsr.BaseClass))
        self.assertTupleEqual(prs_obsr.observer.args, ())
        self.assertDictEqual(prs_obsr.observer.kwargs, {})
        self.assertListEqual(prs_obsr.methods, [])
        self.assertListEqual(prs_obsr.observables, [])
        self.assertIs(prs_obsr.substitute.callee, person)
        self.assertTupleEqual(prs_obsr.substitute.callee_args, ())
        self.assertDictEqual(prs_obsr.substitute.callee_kwargs, {})
        self.assertIs(prs_obsr.__annotations__, person.__annotations__)
        self.assertIs(prs_obsr.__doc__, person.__doc__)
        self.assertIs(prs_obsr.__name__, person.__name__)

        # ---------------------------------------------------------------------
        # Calls
        res.clear()
        prs_obsr()
        self.assertListEqual(res, ["person says 'Hello?'"])
        res.clear()
        dog_obsl()
        self.assertListEqual(res, ["dog acts 'Brrr!'"])

    def test_nodecoration_params(self):
        """Unittest: Functions - no decoration, parameters"""

        # Result list
        res = list()

        # Observable
        def dog(act1: str = "Brrr", act2: str = "!") -> None:
            """A dog function"""
            res.append(f"{dog.__name__} acts '{act1}{act2}'")

        # Observer
        def person(say1: str = "Hello", say2: str = "?") -> None:
            """A person function"""
            res.append(f"{person.__name__} says '{say1}{say2}'")

        # ---------------------------------------------------------------------
        # Test setup: Observable, only
        dog_obsl = Observable(dog, 'Woof, ', act2='Woof!')

        # Checks for dog_obsl
        self.assertTrue(isinstance(dog_obsl, Observable))
        self.assertIs(dog_obsl.BaseClass, BaseObservable)
        self.assertTrue(isinstance(dog_obsl.observable, dog_obsl.BaseClass))
        self.assertTrue(hasattr(dog_obsl.observable, 'register'))
        self.assertTrue(hasattr(dog_obsl.observable, 'unregister'))
        self.assertTrue(hasattr(dog_obsl.observable, 'dispatch'))
        self.assertTrue(hasattr(dog_obsl.observable, 'observers'))
        self.assertTupleEqual(dog_obsl.observable.args, ())
        self.assertDictEqual(dog_obsl.observable.kwargs, {})
        self.assertIs(dog_obsl.activate, Activation.AFTER)
        self.assertListEqual(dog_obsl.methods, [])
        self.assertIs(dog_obsl.substitute.callee, dog)
        self.assertTupleEqual(dog_obsl.substitute.callee_args, ('Woof, ',))
        self.assertDictEqual(dog_obsl.substitute.callee_kwargs,
                             {'act2': 'Woof!'})
        self.assertIs(dog_obsl.__annotations__, dog.__annotations__)
        self.assertIs(dog_obsl.__doc__, dog.__doc__)
        self.assertIs(dog_obsl.__name__, dog.__name__)

        # ---------------------------------------------------------------------
        # Calls
        res.clear()
        person()
        self.assertListEqual(res, ["person says 'Hello?'"])
        res.clear()
        dog_obsl()
        self.assertListEqual(res, ["dog acts 'Woof, Woof!'"])

        # ---------------------------------------------------------------------
        # Test setup: Observable & Observer
        prs_obsr = Observer(person, 'Ooops', say2='!')

        # Checks for prs_obsr
        self.assertTrue(isinstance(prs_obsr, Observer))
        self.assertIs(prs_obsr.BaseClass, BaseObserver)
        self.assertTrue(isinstance(prs_obsr.observer, prs_obsr.BaseClass))
        self.assertTupleEqual(prs_obsr.observer.args, ())
        self.assertDictEqual(prs_obsr.observer.kwargs, {})
        self.assertListEqual(prs_obsr.methods, [])
        self.assertListEqual(prs_obsr.observables, [])
        self.assertIs(prs_obsr.substitute.callee, person)
        self.assertTupleEqual(prs_obsr.substitute.callee_args, ('Ooops',))
        self.assertDictEqual(prs_obsr.substitute.callee_kwargs, {'say2': '!'})
        self.assertIs(prs_obsr.__annotations__, person.__annotations__)
        self.assertIs(prs_obsr.__doc__, person.__doc__)
        self.assertIs(prs_obsr.__name__, person.__name__)

        # ---------------------------------------------------------------------
        # Calls
        res.clear()
        prs_obsr()
        self.assertListEqual(res, ["person says 'Ooops!'"])
        res.clear()
        dog_obsl()
        self.assertListEqual(res, ["dog acts 'Woof, Woof!'"])

    def test_decoration_noparams(self):
        """Unittest: Functions - decoration, no parameters"""

        # Result list
        res = list()

        # Observable
        def dog(act: str = "Brrr!") -> None:
            """A dog function"""
            res.append(f"{dog.__name__} acts '{act}'")

        # Observer
        def person(say: str = "Hello?") -> None:
            """A person function"""
            res.append(f"{person.__name__} says '{say}'")

        # ---------------------------------------------------------------------
        # Test setup: Observable, only
        dog_obsl = Observable(None, 'deco_arg', kw='deco_kwarg')(dog)

        # Checks for dog_obsl
        self.assertTrue(isinstance(dog_obsl, Observable))
        self.assertIs(dog_obsl.BaseClass, BaseObservable)
        self.assertTrue(isinstance(dog_obsl.observable, dog_obsl.BaseClass))
        self.assertTrue(hasattr(dog_obsl.observable, 'register'))
        self.assertTrue(hasattr(dog_obsl.observable, 'unregister'))
        self.assertTrue(hasattr(dog_obsl.observable, 'dispatch'))
        self.assertTrue(hasattr(dog_obsl.observable, 'observers'))
        self.assertTupleEqual(dog_obsl.observable.args, ('deco_arg',))
        self.assertDictEqual(dog_obsl.observable.kwargs, {'kw': 'deco_kwarg'})
        self.assertIs(dog_obsl.activate, Activation.AFTER)
        self.assertListEqual(dog_obsl.methods, [])
        self.assertIs(dog_obsl.substitute.callee, dog)
        self.assertTupleEqual(dog_obsl.substitute.callee_args, ())
        self.assertDictEqual(dog_obsl.substitute.callee_kwargs, {})
        self.assertIs(dog_obsl.__annotations__, dog.__annotations__)
        self.assertIs(dog_obsl.__doc__, dog.__doc__)
        self.assertIs(dog_obsl.__name__, dog.__name__)

        # Calls
        res.clear()
        person()
        self.assertListEqual(res, ["person says 'Hello?'"])
        res.clear()
        dog_obsl()
        self.assertListEqual(res, ["dog acts 'Brrr!'"])

        # ---------------------------------------------------------------------
        # Test setup: Observable & Observer
        prs_obsr = Observer(None, 'deco_arg', kw='deco_kwarg')(person)

        # Checks for prs_obsr
        self.assertTrue(isinstance(prs_obsr, Observer))
        self.assertIs(prs_obsr.BaseClass, BaseObserver)
        self.assertTrue(isinstance(prs_obsr.observer, prs_obsr.BaseClass))
        self.assertTupleEqual(prs_obsr.observer.args, ('deco_arg',))
        self.assertDictEqual(prs_obsr.observer.kwargs, {'kw': 'deco_kwarg'})
        self.assertListEqual(prs_obsr.methods, [])
        self.assertListEqual(prs_obsr.observables, [])
        self.assertIs(prs_obsr.substitute.callee, person)
        self.assertTupleEqual(prs_obsr.substitute.callee_args, ())
        self.assertDictEqual(prs_obsr.substitute.callee_kwargs, {})
        self.assertIs(prs_obsr.__annotations__, person.__annotations__)
        self.assertIs(prs_obsr.__doc__, person.__doc__)
        self.assertIs(prs_obsr.__name__, person.__name__)

        # ---------------------------------------------------------------------
        # Calls
        res.clear()
        prs_obsr()
        self.assertListEqual(res, ["person says 'Hello?'"])
        res.clear()
        dog_obsl()
        self.assertListEqual(res, ["dog acts 'Brrr!'"])

    def test_decoration_params(self):
        """Unittest: Functions - decoration, parameters"""

        # Result list
        res = list()

        # Observable
        def dog(act1: str = "Brrr", act2: str = "!") -> None:
            """A dog function"""
            res.append(f"{dog.__name__} acts '{act1}{act2}'")

        # Observer
        def person(say1: str = "Hello", say2: str = "?") -> None:
            """A person function"""
            res.append(f"{person.__name__} says '{say1}{say2}'")

        # ---------------------------------------------------------------------
        # Test setup: Observable, only
        dog_obsl = Observable(None, 'deco_arg', kw='deco_kwarg')(
            dog, 'Woof, ', act2='Woof!')

        # Checks for dog_obsl
        self.assertTrue(isinstance(dog_obsl, Observable))
        self.assertIs(dog_obsl.BaseClass, BaseObservable)
        self.assertTrue(isinstance(dog_obsl.observable, dog_obsl.BaseClass))
        self.assertTrue(hasattr(dog_obsl.observable, 'register'))
        self.assertTrue(hasattr(dog_obsl.observable, 'unregister'))
        self.assertTrue(hasattr(dog_obsl.observable, 'dispatch'))
        self.assertTrue(hasattr(dog_obsl.observable, 'observers'))
        self.assertTupleEqual(dog_obsl.observable.args, ('deco_arg',))
        self.assertDictEqual(dog_obsl.observable.kwargs, {'kw': 'deco_kwarg'})
        self.assertIs(dog_obsl.activate, Activation.AFTER)
        self.assertListEqual(dog_obsl.methods, [])
        self.assertIs(dog_obsl.substitute.callee, dog)
        self.assertTupleEqual(dog_obsl.substitute.callee_args, ('Woof, ',))
        self.assertDictEqual(dog_obsl.substitute.callee_kwargs,
                             {'act2': 'Woof!'})
        self.assertIs(dog_obsl.__annotations__, dog.__annotations__)
        self.assertIs(dog_obsl.__doc__, dog.__doc__)
        self.assertIs(dog_obsl.__name__, dog.__name__)

        # ---------------------------------------------------------------------
        # Calls
        res.clear()
        person()
        self.assertListEqual(res, ["person says 'Hello?'"])
        res.clear()
        dog_obsl()
        self.assertListEqual(res, ["dog acts 'Woof, Woof!'"])

        # ---------------------------------------------------------------------
        # Test setup: Observable & Observer
        prs_obsr = Observer(None, 'deco_arg', kw='deco_kwarg')(
            person, 'Ooops', say2='!')

        # Checks for prs_obsr
        self.assertTrue(isinstance(prs_obsr, Observer))
        self.assertIs(prs_obsr.BaseClass, BaseObserver)
        self.assertTrue(isinstance(prs_obsr.observer, prs_obsr.BaseClass))
        self.assertTupleEqual(prs_obsr.observer.args, ('deco_arg',))
        self.assertDictEqual(prs_obsr.observer.kwargs, {'kw': 'deco_kwarg'})
        self.assertListEqual(prs_obsr.methods, [])
        self.assertListEqual(prs_obsr.observables, [])
        self.assertIs(prs_obsr.substitute.callee, person)
        self.assertTupleEqual(prs_obsr.substitute.callee_args, ('Ooops',))
        self.assertDictEqual(prs_obsr.substitute.callee_kwargs, {'say2': '!'})
        self.assertIs(prs_obsr.__annotations__, person.__annotations__)
        self.assertIs(prs_obsr.__doc__, person.__doc__)
        self.assertIs(prs_obsr.__name__, person.__name__)

        # ---------------------------------------------------------------------
        # Calls
        res.clear()
        prs_obsr()
        self.assertListEqual(res, ["person says 'Ooops!'"])
        res.clear()
        dog_obsl()
        self.assertListEqual(res, ["dog acts 'Woof, Woof!'"])

    def test_observable_activation(self):
        """Unittest: Functions - Activation"""

        # Result list
        res = list()

        # Observable
        def dog(act: str = "Brrr!") -> None:
            """A dog function"""
            res.append(f"{dog.__name__} acts '{act}'")

        # Observer
        def person(say: str = "Hello?") -> None:
            """A person function"""
            res.append(f"{person.__name__} says '{say}'")

        # ---------------------------------------------------------------------
        # Test
        self.assertEqual(len(Activation), 4)

        self.assertNotEqual(Activation.NONE, Activation.BEFORE)
        self.assertNotEqual(Activation.NONE, Activation.AFTER)
        self.assertNotEqual(Activation.NONE, Activation.BEFORE_AND_AFTER)
        self.assertNotEqual(Activation.BEFORE, Activation.AFTER)
        self.assertNotEqual(Activation.BEFORE, Activation.BEFORE_AND_AFTER)
        self.assertNotEqual(Activation.AFTER, Activation.BEFORE_AND_AFTER)

        self.assertEqual(Activation.BEFORE_AND_AFTER,
                         Activation.BEFORE | Activation.AFTER)

        # ---------------------------------------------------------------------
        # Decoration
        dog_obsl = Observable(dog)
        prs_obsr = Observer(observables=dog_obsl)(person)

        # Person says default
        res.clear()
        prs_obsr()
        self.assertListEqual(res, ["person says 'Hello?'"])

        # ---------------------------------------------------------------------

        # Default activation is AFTER
        self.assertIs(dog_obsl.activate, Activation.AFTER)

        # Activation: NONE
        dog_obsl.activate = Activation.NONE
        res.clear()
        dog_obsl()
        self.assertListEqual(res, ["dog acts 'Brrr!'"])

        # Activation: BEFORE
        dog_obsl.activate = Activation.BEFORE
        res.clear()
        dog_obsl()
        self.assertListEqual(res, ["person says 'Hello?'",
                                   "dog acts 'Brrr!'"])

        # Activation: AFTER
        dog_obsl.activate = Activation.AFTER
        res.clear()
        dog_obsl()
        self.assertListEqual(res, ["dog acts 'Brrr!'",
                                   "person says 'Hello?'"])

        # Activation: BEFORE & AFTER
        dog_obsl.activate = Activation.BEFORE_AND_AFTER
        res.clear()
        dog_obsl()
        self.assertListEqual(res, ["person says 'Hello?'",
                                   "dog acts 'Brrr!'",
                                   "person says 'Hello?'"])

    def test_observable_functions(self):
        """Unittest: Functions - Interface register, unregister, dispatch"""

        # Result list
        res = list()

        # Observable
        def dog(act: str = "Woof!") -> None:
            """A dog function"""
            res.append(f"{dog.__name__} acts '{act}'")

        # Observer
        def person(say: str = "Hello?") -> None:
            """A person function"""
            res.append(f"{person.__name__} says '{say}'")

        # ---------------------------------------------------------------------
        # Test scenario
        def test(Person, Dog):
            # Person registered as an observer of dog by decorations above
            self.assertTrue(cmp(Dog.observable.observers().values(),
                                {F(Person, say="What's up?")}))

            # Repeated registration override current registration
            Dog.observable.register(Person)
            self.assertTrue(cmp(Dog.observable.observers().values(),
                                {F(Person)}))
            Dog.observable.register(Person, say="What's up?")
            self.assertTrue(cmp(Dog.observable.observers().values(),
                                {F(Person, say="What's up?")}))

            # Unregister person as an observer of dog
            Dog.observable.unregister(Person)
            self.assertDictEqual(Dog.observable.observers(classbased=True), {})
            self.assertDictEqual(Dog.observable.observers(), dict())

            # Re-register person as an observer of dog
            Dog.observable.register(Person)
            self.assertTrue(cmp(Dog.observable.observers().values(),
                                {F(Person)}))
            Dog.observable.register(Person, say="What's up?")
            self.assertTrue(cmp(Dog.observable.observers().values(),
                                {F(Person, say="What's up?")}))

            # Unregister *all* observers of dog
            Dog.observable.unregister()
            self.assertDictEqual(Dog.observable.observers(classbased=True), {})
            self.assertDictEqual(Dog.observable.observers(), dict())

            # Re-register person as an observer of dog
            Dog.observable.register(Person)
            self.assertTrue(cmp(Dog.observable.observers().values(),
                                {F(Person)}))
            Dog.observable.register(Person, say="What's up?")
            self.assertTrue(cmp(Dog.observable.observers().values(),
                                {F(Person, say="What's up?")}))

            # Person say default
            res.clear()
            Person()
            self.assertListEqual(res, ["person says 'Hello?'"])

            # Person say customized text using positional parameter
            res.clear()
            Person("Where is my dog?")
            self.assertListEqual(res, ["person says 'Where is my dog?'"])

            # Person say customized text using keyword parameter
            res.clear()
            Person(say="Where is my dog?")
            self.assertListEqual(res, ["person says 'Where is my dog?'"])

            # Dog act default triggers person say decoration default
            res.clear()
            Dog()
            self.assertListEqual(res, ["dog acts 'Woof!'",
                                       "person says 'What's up?'"])

            # Dog act customized text triggers person say decoration default
            res.clear()
            Dog("Brrr!")
            self.assertListEqual(res, ["dog acts 'Brrr!'",
                                       "person says 'What's up?'"])

            # Initiate additional customized dispatch
            res.clear()
            Dog("Brrr!")
            Dog.observable.dispatch(observer=Person, say="Quiet, please!")
            self.assertListEqual(res, ["dog acts 'Brrr!'",
                                       "person says 'What's up?'",
                                       "person says 'Quiet, please!'"])

            # Switch notification off/on
            res.clear()
            Dog.activate = Activation.NONE
            Dog("Brrr!")
            self.assertListEqual(res, ["dog acts 'Brrr!'"])

            res.clear()
            Dog.activate = Activation.BEFORE
            Dog("Brrr!")
            self.assertListEqual(res, ["person says 'What's up?'",
                                       "dog acts 'Brrr!'"])

            res.clear()
            Dog.activate = Activation.AFTER
            Dog("Brrr!")
            self.assertListEqual(res, ["dog acts 'Brrr!'",
                                       "person says 'What's up?'"])

            res.clear()
            Dog.activate = Activation.BEFORE_AND_AFTER
            Dog("Brrr!")
            self.assertListEqual(res, ["person says 'What's up?'",
                                       "dog acts 'Brrr!'",
                                       "person says 'What's up?'"])

            Dog.activate = Activation.AFTER

        # ---------------------------------------------------------------------
        # Test: Decoration via Observable
        prs_obsr = Observer(person)
        dog_obsl = Observable(observers=F(prs_obsr, say="What's up?"))(dog)
        test(prs_obsr, dog_obsl)

        # ---------------------------------------------------------------------
        # Test: Decoration via Observer
        dog_obsl = Observable(dog)
        prs_obsr = Observer(observables=X(dog_obsl, say="What's up?"))(person)
        test(prs_obsr, dog_obsl)

    def test_observable_class_init(self):
        """Unittest: Class - init()"""

        # Result list
        res = list()

        # Observable
        class Dog:
            """A Dog Type"""

            def __init__(self, name: str = "DOG"):
                self.name = name
                res.append(f"{self.name} acts 'INIT'")

            def untouched(self):
                """Untouched by decoration"""
                res.append(f"{self.name} is untouched")

        # Observer
        class Person:
            """A Person Type"""

            def __init__(self, name: str = "PERSON"):
                self.name = name
                res.append(f"{self.name} says 'INIT'")

            def untouched(self):
                """Untouched by decoration"""
                res.append(f"{self.name} is untouched")

        # ---------------------------------------------------------------------
        # Test setup: Decoration
        Prs_obsr = Observer(Person)
        Dog_obsl = Observable(observers=Prs_obsr)(Dog)

        # Person registered as an observer of dog by decorations above
        self.assertDictEqual(Dog_obsl.observable.observers(),
                             {F(Prs_obsr): F(Prs_obsr)})

        # Init a Person
        res.clear()
        p = Prs_obsr()
        self.assertListEqual(res, ["PERSON says 'INIT'"])

        # Init a Dog with Person as an observer
        res.clear()
        d = Dog_obsl()
        self.assertListEqual(res, ["DOG acts 'INIT'", "PERSON says 'INIT'"])

        # The untouchables
        res.clear()
        p.untouched()
        self.assertEqual(res, ['PERSON is untouched'])
        res.clear()
        d.untouched()
        self.assertEqual(res, ['DOG is untouched'])

    def test_observable_class_staticmethod01(self):
        """Unittest: Class - staticmethod()"""

        # Result list
        res = list()

        # Observable
        class Dog:
            """A Dog Type"""

            @staticmethod
            def static_method(act: str = 'Woof!'):
                """To be decorated"""
                res.append(f"Dog act '{act}'")

            @staticmethod
            def static_untouched():
                """Untouched by decoration"""
                res.append(f"Dog is untouched")

        # Observer
        class Person:
            """A Person Type"""

            @staticmethod
            def static_method(say: str = 'Hello?'):
                """To be decorated"""
                res.append(f"Person say '{say}'")

            @staticmethod
            def static_untouched():
                """Untouched by decoration"""
                res.append(f"Person is untouched")

        # ---------------------------------------------------------------------
        # Test: function decoration
        Person.static_method = Observer(Person.static_method)
        Dog.static_method = Observable(
            observers=F(Person.static_method, say="What's up?"))(
            Dog.static_method)

        # Person registered as an observer of dog by decorations above
        self.assertTrue(cmp(Dog.static_method.observable.observers().values(),
                            {F(Person.static_method, say="What's up?")}))

        # Person
        res.clear()
        Person.static_method()
        self.assertListEqual(res, ["Person say 'Hello?'"])

        # Dog with Person as an observer
        res.clear()
        Dog.static_method()
        self.assertListEqual(res, ["Dog act 'Woof!'",
                                   "Person say 'What's up?'"])

        # The untouchables
        res.clear()
        Person.static_untouched()
        self.assertEqual(res, ['Person is untouched'])
        res.clear()
        Dog.static_untouched()
        self.assertEqual(res, ['Dog is untouched'])

    def test_observable_class_staticmethod02(self):
        """Unittest: Class -  staticmethod()"""

        # Result list
        res = list()

        # Observable
        class Dog:
            """A Dog Type"""

            @staticmethod
            def static_method(act: str = 'Woof!'):
                """To be decorated"""
                res.append(f"Dog act '{act}'")

            @staticmethod
            def static_untouched():
                """Untouched by decoration"""
                res.append(f"Dog is untouched")

        # Observer
        class Person:
            """A Person Type"""

            @staticmethod
            def static_method(say: str = 'Hello?'):
                """To be decorated"""
                res.append(f"Person say '{say}'")

            @staticmethod
            def static_untouched():
                """Untouched by decoration"""
                res.append(f"Person is untouched")

        # ---------------------------------------------------------------------
        # Test: class decoration
        Person = Observer(methods=Person.static_method)(Person)
        Dog = Observable(observers=F(Person.static_method, say="What's up?"),
                         methods=Dog.static_method)(Dog)

        # Person registered as an observer of dog by decorations above
        self.assertTrue(cmp(Dog.static_method.observable.observers().values(),
                            {F(Person.static_method, say="What's up?")}))

        # Person
        res.clear()
        Person.static_method()
        self.assertListEqual(res, ["Person say 'Hello?'"])

        # Dog with Person as an observer
        res.clear()
        Dog.static_method()
        self.assertListEqual(res, ["Dog act 'Woof!'",
                                   "Person say 'What's up?'"])

        # The untouchables
        res.clear()
        Person.static_untouched()
        self.assertEqual(res, ['Person is untouched'])
        res.clear()
        Dog.static_untouched()
        self.assertEqual(res, ['Dog is untouched'])

    def test_observable_class_classmethod01(self):
        """Unittest: Class - classmethod()"""

        # Result list
        res = list()

        # Observable
        class Dog:
            """A Dog Type"""

            @classmethod
            def class_method(cls, act: str = 'Woof!'):
                """A classmethod"""
                res.append(f"{cls.__name__} act '{act}'")

        # Observer
        class Person:
            """A Person Type"""

            @classmethod
            def class_method(cls, say: str = 'Hello?'):
                """A classmethod"""
                res.append(f"{cls.__name__} say '{say}'")

        # ---------------------------------------------------------------------
        # Test: function decoration
        Person.class_method = Observer(Person.class_method)
        Dog.class_method = Observable(
            observers=F(Person.class_method, say="What's up?"))(
            Dog.class_method)

        # Person registered as an observer of dog by decorations above
        self.assertTrue(cmp(Dog.class_method.observable.observers().values(),
                            {F(Person.class_method, say="What's up?")}))

        # Person
        res.clear()
        Person.class_method()
        self.assertListEqual(res, ["Person say 'Hello?'"])

        # Dog with Person as an observer
        res.clear()
        Dog.class_method()
        self.assertListEqual(res, ["Dog act 'Woof!'",
                                   "Person say 'What's up?'"])

    def test_observable_class_classmethod02(self):
        """Unittest: Class - classmethod()"""

        # Result list
        res = list()

        # Observable
        class Dog:
            """A Dog Type"""

            @classmethod
            def class_method(cls, act: str = 'Woof!'):
                """A classmethod"""
                res.append(f"{cls.__name__} act '{act}'")

        # Observer
        class Person:
            """A Person Type"""

            @classmethod
            def class_method(cls, say: str = 'Hello?'):
                """A classmethod"""
                res.append(f"{cls.__name__} say '{say}'")

        # ---------------------------------------------------------------------
        # Test: class decoration
        Person = Observer(methods=Person.class_method)(Person)
        Dog = Observable(observers=F(Person.class_method, say="What's up?"),
                         methods=Dog.class_method)(Dog)

        # Person registered as an observer of dog by decorations above
        self.assertTrue(cmp(Dog.class_method.observable.observers().values(),
                            {F(Person.class_method, say="What's up?")}))

        # Person
        res.clear()
        Person.class_method()
        self.assertListEqual(res, ["Person say 'Hello?'"])

        # Dog with Person as an observer
        res.clear()
        Dog.class_method()
        self.assertListEqual(res, ["Dog act 'Woof!'",
                                   "Person say 'What's up?'"])

    def test_observable_class_objectmethod01(self):
        """Unittest: Class - object method()"""

        # Result list
        res = list()

        # Observable
        class Dog:
            """A Dog Type"""

            def object_method(self, act: str = 'Woof!'):
                """An objectmethod"""
                res.append(f"{self.__class__.__name__} act '{act}'")

        # Observer
        class Person:
            """A Person Type"""

            def object_method(self, say: str = 'Hello?'):
                """An objectmethod"""
                res.append(f"{self.__class__.__name__} say '{say}'")

        # ---------------------------------------------------------------------
        # Test: function decoration
        Per_obsr = Person()
        Dog_obsl = Dog()

        Per_obsr.object_method = Observer(Per_obsr.object_method)
        Dog_obsl.object_method = Observable(
            observers=F(Per_obsr.object_method, say="What's up?"))(
            Dog_obsl.object_method)

        # Person registered as an observer of dog by decorations above
        self.assertTrue(
            cmp(Dog_obsl.object_method.observable.observers().values(),
                {F(Per_obsr.object_method, say="What's up?")}))

        # Person
        res.clear()
        Per_obsr.object_method()
        self.assertListEqual(res, ["Person say 'Hello?'"])

        # Dog with Person as an observer
        res.clear()
        Dog_obsl.object_method()
        self.assertListEqual(res, ["Dog act 'Woof!'",
                                   "Person say 'What's up?'"])

    def test_observable_class_objectmethod02(self):
        """Unittest: Class - object method()"""

        # Result list
        res = list()

        # Observable
        class Dog:
            """A Dog Type"""

            def object_method(self, act: str = 'Woof!'):
                """An objectmethod"""
                res.append(f"{self.__class__.__name__} act '{act}'")

        # Observer
        class Person:
            """A Person Type"""

            def object_method(self, say: str = 'Hello?'):
                """An objectmethod"""
                res.append(f"{self.__class__.__name__} say '{say}'")

        # ---------------------------------------------------------------------
        # Test: object decoration
        Per_obsr = Person()
        Dog_obsl = Dog()

        Per_obsr = Observer(methods=X('object_method'))(Per_obsr)
        Dog_obsl = Observable(
            observers=F(Per_obsr.object_method, say="What's up?"),
            methods=F('object_method'))(Dog_obsl)

        # Person registered as an observer of dog by decorations above
        self.assertTrue(
            cmp(Dog_obsl.object_method.observable.observers().values(),
                {F(Per_obsr.object_method, say="What's up?")}))

        # Person
        res.clear()
        Per_obsr.object_method()
        self.assertListEqual(res, ["Person say 'Hello?'"])

        # Dog with Person as an observer
        res.clear()
        Dog_obsl.object_method()
        self.assertListEqual(res, ["Dog act 'Woof!'",
                                   "Person say 'What's up?'"])


# -----------------------------------------------------------------------------
# Execution
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding=UTF-8 -*-
# vim: fileencoding=UTF-8 tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -----------------------------------------------------------------------------
# Document Description
"""**Banner**"""

# -----------------------------------------------------------------------------
# Module Level Dunders
__title__ = "Banner"
__module__ = "banner.py"
__author__ = "Martin Abel"
__maintainer__ = "Martin Abel"
__credits__ = ["Martin Abel"]
__company__ = "eVation"
__email__ = "python@evation.eu"
__url__ = "http://evation.eu"
__copyright__ = f"(c) copyright 2020-2023, {__company__}"
__created__ = "2020-01-01"
__version__ = "0.9.0.12"
__date__ = "2023-06-23"
__time__ = "08:53:07"
__state__ = "Beta"
__license__ = "PSF"

__all__ = ["__banner"]

# -----------------------------------------------------------------------------
# Libraries & Modules
import unittest
import sys

from os.path import dirname
from importlib import import_module
from decoratory.__main__ import get_file_location


# -----------------------------------------------------------------------------
# Banner Display
def __banner(title, version, date, time, docs,
             author: str = __author__,
             maintainer: str = __maintainer__,
             company: str = __company__,
             email: str = __email__,
             url: str = __url__,
             copyright: str = __copyright__,
             state: str = __state__,
             license: str = __license__,
             file_license: str = "License.txt"):
    """**Banner**

    Banner is an auxiliary module for displaying package and module
    information.
    """
    module = title.lower()
    name = title.capitalize()

    arg = sys.argv[1].strip().lower() if len(sys.argv) > 1 else "message"
    if arg in "--version":
        print(f"{module}: Version {version}, "
              f"Build {date} {time}, State '{state}'")
    elif arg in "--copyright":
        print(f"{copyright}")
    elif arg in "--license" or arg in "--licence":
        print(f"\n{'-' * 79}")
        print(f"{title} ships under the license: {license}".center(79))
        print(f"{'-' * 79}")
        file_path = get_file_location(file_license)
        if file_path is not None:
            with open(file_path, "r") as fp:
                txt = fp.read()
            print("\n" + txt)
        else:
            print("*** No additional data accessible. Please look for "
                  f"file '{file_license}' ***".center(79, " "))
        # print(f"{'-' * 79}")
        # print(f"File: {file_path}".center(79, " "))
        print(f"{'-' * 79}")
    elif arg in "--documentation":
        print()
        for doc in docs:
            print(f"{' ' + doc.__name__ + ' ':-^79s}\n")
            print(doc.__doc__)  # print(help(doc))
        print(f"{'-' * 79}")
    elif arg in "--information":
        print(f"\nPeople and contact information for module {module}:\n")
        print(f"Author    : {author}")
        print(f"Maintainer: {maintainer}")
        print(f"Company   : {company}")
        print(f"Email     : {email}")
        print(f"Web       : {url}")
    elif arg in "--test":
        try:
            print(f"\n{'-' * 70}")
            print(f" Running unit test for module: {module} ".center(70, '-'))
            print(f"{'-' * 70}\n")
            sys_path = dirname(get_file_location(f"test_{module}.py"))
            if sys_path and sys_path not in sys.path:
                sys.path.insert(0, sys_path)
            loader = unittest.TestLoader().loadTestsFromModule
            mdl = import_module(f"test_{module}")
            suite = loader(mdl)
            unittest.TextTestRunner(verbosity=2).run(suite)
        except (ModuleNotFoundError, ImportError, NameError, Exception):
            print(f"*** No unit test available for module "
                  f"{module} ***\n".center(70, " "))
        print(f"{'-' * 70}")
    else:
        print(f"""
-------------------------------------------------------------------------------
{title.upper(): ^79}
-------------------------------------------------------------------------------

{module} is a Python module you cannot execute directly from command line.
This module provides a {name} decorator.

{module} is provided by {company} under the {license} license included with 
this module. It comes without any warranty, but with the intention of saving 
working time and improving code quality and productivity.

All kinds of bugs, questions, improvement suggestions, change requests etc. 
are welcome to be directed to the product maintainers email {email}.

The current version of the installed package is:
{module}: Version {version}, Build {date} {time}, State '{state}'

Syntax:   python -m decoratory.{module} [option]

Possible options are:

  -h, --help           show this help message
  -v, --version        show module {module} version information      
  -c, --copyright      show module {module} copyright statement
  -l, --license        show module {module} license statement (License.txt)
      --licence        an alias to --license
  -d, --documentation  display selected module code documentation
  -t, --test           run unit test for module {module}
  -i, --info           people and contact information

-------------------------------------------------------------------------------
{copyright: <69s}{date: >10s}
-------------------------------------------------------------------------------
""")


# -----------------------------------------------------------------------------
# Entry Point
if __name__ == '__main__':
    __banner(title=__title__,
             version=__version__,
             date=__date__,
             time=__time__,
             docs=(__banner,))



#!/usr/bin/env python
# -*- coding=UTF-8 -*-
# vim: fileencoding=UTF-8 tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -----------------------------------------------------------------------------
# Document Description
"""**Test Wrapper**"""

# -----------------------------------------------------------------------------
# Module Level Dunders
__title__ = "Test Wrapper"
__module__ = "test_wrapper.py"
__author__ = "Martin Abel"
__maintainer__ = "Martin Abel"
__credits__ = ["Martin Abel"]
__company__ = "eVation"
__email__ = "python@evation.eu"
__url__ = "http://evation.eu"
__copyright__ = f"(c) copyright 2020-2023, {__company__}"
__created__ = "2020-01-01"
__version__ = "0.9.0.1"
__date__ = "2023-06-21"
__time__ = "13:37:48"
__state__ = "Beta"
__license__ = "PSF"

# -----------------------------------------------------------------------------
# Libraries & Modules
import unittest

from decoratory.wrapper import Wrapper
from decoratory.basic import F


# -----------------------------------------------------------------------------
# Test Class
# noinspection PyPep8Naming
class TestWrapper(unittest.TestCase):

    def setUp(self):
        """Preparation"""
        pass

    def tearDown(self):
        """Wrap-up"""
        pass

    def test_decoration_customized(self):
        """Unittest: Wrapper with customized default text, only"""

        # Result list
        res = list()

        def printer(txt: str = "Default text"):
            """The default printer"""
            res.append(f"{printer.__name__}: '{txt}'")

        wrp_printer = Wrapper(printer, "Customized text")

        # ---------------------------------------------------------------------
        # Checks
        self.assertTrue(isinstance(wrp_printer, Wrapper))

        # Call
        res.clear()
        wrp_printer()
        self.assertListEqual(res, ["printer: 'Customized text'"])

        wrp_printer = Wrapper(printer, txt="Customized phrase")

        res.clear()
        wrp_printer()
        self.assertListEqual(res, ["printer: 'Customized phrase'"])

    def test_decoration_nobrackets(self):
        """Unittest: Wrapper without brackets"""

        # Result list
        res = list()

        @Wrapper
        def printer(txt: str = "Default text"):
            """The default printer"""
            res.append(f"{printer.__name__}: '{txt}'")

        # ---------------------------------------------------------------------
        # Checks
        self.assertTrue(isinstance(printer, Wrapper))

        # Call
        res.clear()
        printer()
        self.assertListEqual(res, ["printer: 'Default text'"])

    def test_decoration_empty_brackets(self):
        """Unittest: Wrapper with empty brackets"""

        # Result list
        res = list()

        @Wrapper()
        def printer(txt: str = "Default text"):
            """The default printer"""
            res.append(f"{printer.__name__}: '{txt}'")

        # ---------------------------------------------------------------------
        # Checks
        self.assertTrue(isinstance(printer, Wrapper))

        # Call
        res.clear()
        printer()
        self.assertListEqual(res, ["printer: 'Default text'"])

    def test_decoration_before_after(self):
        """Unittest: Wrapper with before and after activities"""

        # Result list
        res = list()

        def before_printer(txt: str = "BEFORE"):
            """The before printer"""
            res.append(f"{before_printer.__name__}: '{txt}'")

        def after_printer(txt: str = "AFTER"):
            """The after printer"""
            res.append(f"{after_printer.__name__}: '{txt}'")

        @Wrapper(before=before_printer, after=after_printer)
        def printer(txt: str = "Default text"):
            """The default printer"""
            res.append(f"{printer.__name__}: '{txt}'")

        # ---------------------------------------------------------------------
        # Checks
        self.assertTrue(isinstance(printer, Wrapper))

        # Call
        res.clear()
        printer()
        self.assertListEqual(res, ["before_printer: 'BEFORE'",
                                   "printer: 'Default text'",
                                   "after_printer: 'AFTER'"])

    def test_decoration_replace(self):
        """Unittest: Wrapper with replace activity"""

        # Result list
        res = list()

        def replace_printer(txt: str = "REPLACE"):
            """The replace printer"""
            res.append(f"{replace_printer.__name__}: '{txt}'")

        @Wrapper(replace=replace_printer)
        def printer(txt: str = "Default text"):
            """The default printer"""
            res.append(f"{printer.__name__}: '{txt}'")

        # ---------------------------------------------------------------------
        # Checks
        self.assertTrue(isinstance(printer, Wrapper))

        # Call
        res.clear()
        printer()
        self.assertListEqual(res, ["replace_printer: 'REPLACE'"])

    def test_decoration_replace_wrapper(self):
        """Unittest: Wrapper with replace wrapper"""

        # Result list
        res = list()

        def replace_wrapper(value: str = "replace"):
            """The replace wrapper"""
            # 1. Edit the call parameters for the original functionality
            value = value.upper()
            # 2. Execute original functionality with modified call parameters
            result = printer.substitute.callee(value)  # (1)
            # 3. Edit the result and return this modified result
            return f"result: '{result}'"

        @Wrapper(replace=replace_wrapper)
        def printer(txt: str = "Default text"):
            """The default printer"""
            res.append(f"{printer.__name__}: '{txt}'")
            return txt

        # ---------------------------------------------------------------------
        # Checks
        self.assertTrue(isinstance(printer, Wrapper))

        # Call
        res.clear()
        res.append(printer())
        self.assertListEqual(res, ["printer: 'REPLACE'",
                                   "result: 'REPLACE'"])

    def test_decoration_before_after_lists(self):
        """Unittest: Wrapper with list of before and after activities"""

        # Result list
        res = list()

        def before_printer(txt: str = "Before text"):
            """The before printer"""
            res.append(f"{before_printer.__name__:14s}: '{txt}'")

        def after_printer(txt: str = "After text"):
            """The after printer"""
            res.append(f"{after_printer.__name__:14s}: '{txt}'")

        @Wrapper(before=[F(before_printer, "Text01"),
                         F(before_printer, txt="Text02"),
                         before_printer],
                 after=[after_printer,
                        F(after_printer, "Text01"),
                        F(after_printer, txt="Text02")])
        def printer(txt: str = "Default text"):
            """The default printer"""
            res.append(f"{printer.__name__:14s}: '{txt}'")

        # ---------------------------------------------------------------------
        # Checks
        self.assertTrue(isinstance(printer, Wrapper))

        # Call
        res.clear()
        printer()
        self.assertListEqual(res, ["before_printer: 'Text01'",
                                   "before_printer: 'Text02'",
                                   "before_printer: 'Before text'",
                                   "printer       : 'Default text'",
                                   "after_printer : 'After text'",
                                   "after_printer : 'Text01'",
                                   "after_printer : 'Text02'"])

    def test_decoration_before_replace_after_lists(self):
        """Unittest: Wrapper with list of before, after and replace activities"""

        # Result list
        res = list()

        def before_printer(txt: str = "Before text"):
            """The before printer"""
            res.append(f"{before_printer.__name__:15s}: '{txt}'")

        def after_printer(txt: str = "After text"):
            """The after printer"""
            res.append(f"{after_printer.__name__:15s}: '{txt}'")

        def replace_printer(add: int = 1, *, result=None):
            """The replace printer"""
            result = add + result if isinstance(result, int) else 0
            res.append(f"{replace_printer.__name__:15s}: result = {result}")
            return result

        @Wrapper(before=[F(before_printer, "First text"),
                         before_printer],
                 after=[after_printer,
                        F(after_printer, "Last text")],
                 replace=[F(replace_printer, 1, result=0),
                          F(replace_printer, 2),
                          F(replace_printer, 3),
                          F(replace_printer, 4)])
        def printer(txt: str = "Default text"):
            """The default printer"""
            res.append(f"{printer.__name__:14s}: '{txt}'")

        # ---------------------------------------------------------------------
        # Checks
        self.assertTrue(isinstance(printer, Wrapper))

        # Call
        res.clear()
        printer()
        self.assertListEqual(res, ["before_printer : 'First text'",
                                   "before_printer : 'Before text'",
                                   "replace_printer: result = 1",
                                   "replace_printer: result = 3",
                                   "replace_printer: result = 6",
                                   "replace_printer: result = 10",
                                   "after_printer  : 'After text'",
                                   "after_printer  : 'Last text'"])


# -----------------------------------------------------------------------------
# Execution
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding=UTF-8 -*-
# vim: fileencoding=UTF-8 tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -----------------------------------------------------------------------------
# Document Description
"""**Test_Basic**"""

# -----------------------------------------------------------------------------
# Module Level Dunders
__title__ = "Test_Basic"
__module__ = "test_basic.py"
__author__ = "Martin Abel"
__maintainer__ = "Martin Abel"
__credits__ = ["Martin Abel"]
__company__ = "eVation"
__email__ = "python@evation.eu"
__url__ = "http://evation.eu"
__copyright__ = f"(c) copyright 2020-2023, {__company__}"
__created__ = "2020-01-01"
__version__ = "0.9.0.1"
__date__ = "2023-06-21"
__time__ = "18:42:36"
__state__ = "Beta"
__license__ = "PSF"

# -----------------------------------------------------------------------------
# Libraries & Modules
import unittest

from decoratory.basic import (Activation, F, Parser, X)


# -----------------------------------------------------------------------------
# Test functions
def cmp(f: F, t: tuple):
    """Compare F with tuple"""
    return (x == y for x, y in zip(f, t))


# -----------------------------------------------------------------------------
# Test Class
# noinspection PyPep8Naming,PyArgumentEqualDefault
class TestDecotools(unittest.TestCase):

    def setUp(self):
        """Preparation"""
        pass

    def tearDown(self):
        """Wrap-up"""
        pass

    def test_activation(self):
        """Unittest: Activation"""

        # ---------------------------------------------------------------------
        # Test
        self.assertEqual(len(Activation), 4)

        self.assertNotEqual(Activation.NONE, Activation.BEFORE)
        self.assertNotEqual(Activation.NONE, Activation.AFTER)
        self.assertNotEqual(Activation.NONE, Activation.BEFORE_AND_AFTER)
        self.assertNotEqual(Activation.BEFORE, Activation.AFTER)
        self.assertNotEqual(Activation.BEFORE, Activation.BEFORE_AND_AFTER)
        self.assertNotEqual(Activation.AFTER, Activation.BEFORE_AND_AFTER)

        self.assertEqual(Activation.BEFORE_AND_AFTER,
                         Activation.BEFORE | Activation.AFTER)

    def test_F(self):
        """Unittest: F wrapper"""

        # A callable
        def function(x, y):
            """Test function"""
            return x + y

        # ---------------------------------------------------------------------
        # Test

        # Callee is mandatory
        try:
            # noinspection PyArgumentList
            F()
        except TypeError:
            pass
        else:
            raise AssertionError

        # Empty parameters
        self.assertTrue(cmp(F(function), (function, (), {})))
        self.assertTrue(cmp(F('function'), ('function', (), {})))

        # Positional and keyword parameters
        self.assertTrue(cmp(F(function, 2, 3),
                            (function, (2, 3), {})))
        self.assertTrue(cmp(F(function, x=2, y=3),
                            (function, (), {'x': 2, 'y': 3})))
        self.assertTrue(cmp(F(function, 2, y=3),
                            (function, (2,), {'y': 3})))

        # repr()
        self.assertEqual(repr(F(function)),
                         f"F({function.__name__})")
        self.assertEqual(repr(F(function, 2, 3)),
                         f"F({function.__name__}, 2, 3)")
        self.assertEqual(repr(F(function, 2, y=3)),
                         f"F({function.__name__}, 2, y=3)")
        self.assertEqual(repr(F(function, x=2, y=3)),
                         f"F({function.__name__}, x=2, y=3)")
        self.assertEqual(repr(F(function, '2', '3')),
                         f'F({function.__name__}, "2", "3")')
        self.assertEqual(repr(F(function, '2', y='3')),
                         f'F({function.__name__}, "2", y="3")')
        self.assertEqual(repr(F(function, x='2', y='3')),
                         f'F({function.__name__}, x="2", y="3")')

        self.assertEqual(repr(F('function')),
                         f'F("function")')
        self.assertEqual(repr(F('function', 2, 3)),
                         f'F("function", 2, 3)')
        self.assertEqual(repr(F('function', 2, y=3)),
                         f'F("function", 2, y=3)')
        self.assertEqual(repr(F('function', x=2, y=3)),
                         f'F("function", x=2, y=3)')
        self.assertEqual(repr(F('function', '2', '3')),
                         f'F("function", "2", "3")')
        self.assertEqual(repr(F('function', '2', y='3')),
                         f'F("function", "2", y="3")')
        self.assertEqual(repr(F('function', x='2', y='3')),
                         f'F("function", x="2", y="3")')

        # str()
        self.assertEqual(str(F(function)),
                         f"{function.__name__}()")
        self.assertEqual(str(F(function, 2, 3)),
                         f"{function.__name__}(2, 3)")
        self.assertEqual(str(F(function, 2, y=3)),
                         f"{function.__name__}(2, y=3)")
        self.assertEqual(str(F(function, x=2, y=3)),
                         f"{function.__name__}(x=2, y=3)")
        self.assertEqual(str(F(function, '2', '3')),
                         f'{function.__name__}("2", "3")')
        self.assertEqual(str(F(function, '2', y='3')),
                         f'{function.__name__}("2", y="3")')
        self.assertEqual(str(F(function, x='2', y='3')),
                         f'{function.__name__}(x="2", y="3")')

        self.assertEqual(str(F('function')),
                         f'function()')
        self.assertEqual(str(F('function', 2, 3)),
                         f'function(2, 3)')
        self.assertEqual(str(F('function', 2, y=3)),
                         f'function(2, y=3)')
        self.assertEqual(str(F('function', x=2, y=3)),
                         f'function(x=2, y=3)')
        self.assertEqual(str(F('function', "2", "3")),
                         f'function("2", "3")')
        self.assertEqual(str(F('function', "2", y="3")),
                         f'function("2", y="3")')
        self.assertEqual(str(F('function', x="2", y="3")),
                         f'function(x="2", y="3")')

        # hash()
        self.assertEqual(hash(F(function)), hash(function))
        self.assertEqual(hash(F(function, 2, 3)), hash(function))
        self.assertEqual(hash(F(function, 2, y=3)), hash(function))
        self.assertEqual(hash(F(function, x=2, y=3)), hash(function))

        self.assertEqual(hash(F('function')), hash('function'))
        self.assertEqual(hash(F('function', 2, 3)), hash('function'))
        self.assertEqual(hash(F('function', 2, y=3)), hash('function'))
        self.assertEqual(hash(F('function', x=2, y=3)), hash('function'))

        self.assertNotEqual(hash(F('function')), hash(function))
        self.assertNotEqual(hash(F('function', 2, 3)), hash(function))
        self.assertNotEqual(hash(F('function', 2, y=3)), hash(function))
        self.assertNotEqual(hash(F('function', x=2, y=3)), hash(function))

        self.assertNotEqual(hash(F(function)), hash('function'))
        self.assertNotEqual(hash(F(function, 2, 3)), hash('function'))
        self.assertNotEqual(hash(F(function, 2, y=3)), hash('function'))
        self.assertNotEqual(hash(F(function, x=2, y=3)), hash('function'))

        # eq()
        self.assertTrue(F(function) == F(function))
        self.assertTrue(F(function) == function)
        self.assertFalse(F('function') == function)
        self.assertFalse(F(function) == 'function')
        self.assertFalse(F(function) == 13)

        # iter()
        itr = iter(F(function, 2, y=3))
        self.assertEqual(next(itr), function)
        self.assertEqual(next(itr), (2,))
        self.assertEqual(next(itr), {'y': 3})

        # eval()
        self.assertEqual(F(function, 2, 3).eval(), 5)
        self.assertEqual(F(function, 2, y=3).eval(), 5)
        self.assertEqual(F(function, x=2, y=3).eval(), 5)

        class A:
            """Test class"""

            @staticmethod
            def stc_method(x, y):
                """Static method"""
                return x + y

        self.assertEqual(F(A.stc_method, 2, 3).eval(), 5)
        self.assertEqual(F(A.stc_method, 2, y=3).eval(), 5)
        self.assertEqual(F(A.stc_method, x=2, y=3).eval(), 5)

        self.assertEqual(F('stc_method', A, 2, 3).eval(), 5)
        self.assertEqual(F('stc_method', A, 2, y=3).eval(), 5)
        self.assertEqual(F('stc_method', A, x=2, y=3).eval(), 5)

        self.assertEqual(F('stc_method', 2, 3).eval(obj=A), 5)
        self.assertEqual(F('stc_method', 2, y=3).eval(obj=A), 5)
        self.assertEqual(F('stc_method', x=2, y=3).eval(obj=A), 5)

        class A:
            """Test class"""

            @classmethod
            def cls_method(cls, x, y):
                """Class method"""
                return x + y

        self.assertEqual(F(A.cls_method, 2, 3).eval(), 5)
        self.assertEqual(F(A.cls_method, 2, y=3).eval(), 5)
        self.assertEqual(F(A.cls_method, x=2, y=3).eval(), 5)

        self.assertEqual(F('cls_method', A, 2, 3).eval(), 5)
        self.assertEqual(F('cls_method', A, 2, y=3).eval(), 5)
        self.assertEqual(F('cls_method', A, x=2, y=3).eval(), 5)

        self.assertEqual(F('cls_method', 2, 3).eval(obj=A), 5)
        self.assertEqual(F('cls_method', 2, y=3).eval(obj=A), 5)
        self.assertEqual(F('cls_method', x=2, y=3).eval(obj=A), 5)

        class A:
            """Test class"""

            def __init__(self, x=2):
                self.x = x

            def obj_method(self, y):
                """Instance method"""
                return self.x + y

        self.assertEqual(F(A(2).obj_method, 3).eval(), 5)
        self.assertEqual(F(A(2).obj_method, y=3).eval(), 5)
        self.assertEqual(F(A(x=2).obj_method, 3).eval(), 5)
        self.assertEqual(F(A(x=2).obj_method, y=3).eval(), 5)

        self.assertEqual(F('obj_method', A(2), 3).eval(), 5)
        self.assertEqual(F('obj_method', A(2), y=3).eval(), 5)
        self.assertEqual(F('obj_method', A(x=2), 3).eval(), 5)
        self.assertEqual(F('obj_method', A(x=2), y=3).eval(), 5)

        self.assertEqual(F('obj_method', 3).eval(obj=A(2)), 5)
        self.assertEqual(F('obj_method', y=3).eval(obj=A(2)), 5)
        self.assertEqual(F('obj_method', 3).eval(obj=A(x=2)), 5)
        self.assertEqual(F('obj_method', y=3).eval(obj=A(x=2)), 5)

        class A:
            """Test class"""

            def __init__(self, x=2):
                self.x = x

            def __call__(self, y):
                return self.x + y

        self.assertEqual(F(A(2), 3).eval(), 5)
        self.assertEqual(F(A(2), y=3).eval(), 5)
        self.assertEqual(F(A(x=2), 3).eval(), 5)
        self.assertEqual(F(A(x=2), y=3).eval(), 5)

        # Setter
        f = F(function)
        try:
            # noinspection PyPropertyAccess
            f.callee = function
        except AttributeError:
            pass
        else:
            raise AttributeError

        f.callee_args = (2, 3, 4)
        self.assertTupleEqual(f.callee_args, (2, 3, 4))
        f.callee_kwargs = {'x': 2, 'y': 3, 'z': 4}
        self.assertDictEqual(f.callee_kwargs, {'x': 2, 'y': 3, 'z': 4})

    def test_X(self):
        """Unittest: X wrapper"""

        # A callable
        def function(x, y):
            """Test function"""
            return int(x) + int(y)

        # ---------------------------------------------------------------------
        # Test

        # Callee is mandatory
        try:
            # noinspection PyArgumentList
            X()
        except TypeError:
            pass
        else:
            raise AssertionError

        # Empty parameters
        self.assertTrue(cmp(X(function), (function, (), {})))
        self.assertTrue(cmp(X('function'), ('function', (), {})))

        # Positional and keyword parameters
        self.assertTrue(cmp(X(function, 2, 3),
                            (function, (2, 3), {})))
        self.assertTrue(cmp(X(function, x=2, y=3),
                            (function, (), {'x': 2, 'y': 3})))
        self.assertTrue(cmp(X(function, 2, y=3),
                            (function, (2,), {'y': 3})))

        # repr()
        self.assertEqual(repr(X(function)),
                         f"X({function.__name__})")
        self.assertEqual(repr(X(function, 2, 3)),
                         f"X({function.__name__}, 2, 3)")
        self.assertEqual(repr(X(function, 2, y=3)),
                         f"X({function.__name__}, 2, y=3)")
        self.assertEqual(repr(X(function, x=2, y=3)),
                         f"X({function.__name__}, x=2, y=3)")
        self.assertEqual(repr(X(function, '2', '3')),
                         f'X({function.__name__}, "2", "3")')
        self.assertEqual(repr(X(function, '2', y='3')),
                         f'X({function.__name__}, "2", y="3")')
        self.assertEqual(repr(X(function, x='2', y='3')),
                         f'X({function.__name__}, x="2", y="3")')

        self.assertEqual(repr(X('function')),
                         f'X("function")')
        self.assertEqual(repr(X('function', 2, 3)),
                         f'X("function", 2, 3)')
        self.assertEqual(repr(X('function', 2, y=3)),
                         f'X("function", 2, y=3)')
        self.assertEqual(repr(X('function', x=2, y=3)),
                         f'X("function", x=2, y=3)')
        self.assertEqual(repr(X('function', '2', '3')),
                         f'X("function", "2", "3")')
        self.assertEqual(repr(X('function', '2', y='3')),
                         f'X("function", "2", y="3")')
        self.assertEqual(repr(X('function', x='2', y='3')),
                         f'X("function", x="2", y="3")')

        # str()
        self.assertEqual(str(X(function)),
                         f"{function.__name__}()")
        self.assertEqual(str(X(function, 2, 3)),
                         f"{function.__name__}(2, 3)")
        self.assertEqual(str(X(function, 2, y=3)),
                         f"{function.__name__}(2, y=3)")
        self.assertEqual(str(X(function, x=2, y=3)),
                         f"{function.__name__}(x=2, y=3)")
        self.assertEqual(str(X(function, '2', '3')),
                         f'{function.__name__}("2", "3")')
        self.assertEqual(str(X(function, '2', y='3')),
                         f'{function.__name__}("2", y="3")')
        self.assertEqual(str(X(function, x='2', y='3')),
                         f'{function.__name__}(x="2", y="3")')

        self.assertEqual(str(X('function')),
                         f'function()')
        self.assertEqual(str(X('function', 2, 3)),
                         f'function(2, 3)')
        self.assertEqual(str(X('function', 2, y=3)),
                         f'function(2, y=3)')
        self.assertEqual(str(X('function', x=2, y=3)),
                         f'function(x=2, y=3)')
        self.assertEqual(str(X('function', "2", "3")),
                         f'function("2", "3")')
        self.assertEqual(str(X('function', "2", y="3")),
                         f'function("2", y="3")')
        self.assertEqual(str(X('function', x="2", y="3")),
                         f'function(x="2", y="3")')

        # hash()
        self.assertEqual(hash(X(function)), hash(function))
        self.assertEqual(hash(X(function, 2, 3)), hash(function))
        self.assertEqual(hash(X(function, 2, y=3)), hash(function))
        self.assertEqual(hash(X(function, x=2, y=3)), hash(function))

        self.assertEqual(hash(X('function')), hash('function'))
        self.assertEqual(hash(X('function', 2, 3)), hash('function'))
        self.assertEqual(hash(X('function', 2, y=3)), hash('function'))
        self.assertEqual(hash(X('function', x=2, y=3)), hash('function'))

        self.assertNotEqual(hash(X('function')), hash(function))
        self.assertNotEqual(hash(X('function', 2, 3)), hash(function))
        self.assertNotEqual(hash(X('function', 2, y=3)), hash(function))
        self.assertNotEqual(hash(X('function', x=2, y=3)), hash(function))

        self.assertNotEqual(hash(X(function)), hash('function'))
        self.assertNotEqual(hash(X(function, 2, 3)), hash('function'))
        self.assertNotEqual(hash(X(function, 2, y=3)), hash('function'))
        self.assertNotEqual(hash(X(function, x=2, y=3)), hash('function'))

        # eq()
        self.assertTrue(X(function) == X(function))
        self.assertTrue(X(function) == function)
        self.assertFalse(X('function') == function)
        self.assertFalse(X(function) == 'function')
        self.assertFalse(X(function) == 13)

        # iter()
        itr = iter(X(function, 2, y=3))
        self.assertEqual(next(itr), function)
        self.assertEqual(next(itr), (2,))
        self.assertEqual(next(itr), {'y': 3})

        # eval()
        self.assertEqual(X(function, 2, 3).eval(), 5)
        self.assertEqual(X(function, 2, y=3).eval(), 5)
        self.assertEqual(X(function, x=2, y=3).eval(), 5)

        class A:
            """Test class"""

            @staticmethod
            def stc_method(x, y):
                """Static method"""
                return x + y

        self.assertEqual(X(A.stc_method, 2, 3).eval(), 5)
        self.assertEqual(X(A.stc_method, 2, y=3).eval(), 5)
        self.assertEqual(X(A.stc_method, x=2, y=3).eval(), 5)

        self.assertEqual(X('stc_method', A, 2, 3).eval(), 5)
        self.assertEqual(X('stc_method', A, 2, y=3).eval(), 5)
        self.assertEqual(X('stc_method', A, x=2, y=3).eval(), 5)

        self.assertEqual(X('stc_method', 2, 3).eval(obj=A), 5)
        self.assertEqual(X('stc_method', 2, y=3).eval(obj=A), 5)
        self.assertEqual(X('stc_method', x=2, y=3).eval(obj=A), 5)

        class A:
            """Test class"""

            @classmethod
            def cls_method(cls, x, y):
                """Class method"""
                return x + y

        self.assertEqual(X(A.cls_method, 2, 3).eval(), 5)
        self.assertEqual(X(A.cls_method, 2, y=3).eval(), 5)
        self.assertEqual(X(A.cls_method, x=2, y=3).eval(), 5)

        self.assertEqual(X('cls_method', A, 2, 3).eval(), 5)
        self.assertEqual(X('cls_method', A, 2, y=3).eval(), 5)
        self.assertEqual(X('cls_method', A, x=2, y=3).eval(), 5)

        self.assertEqual(X('cls_method', 2, 3).eval(obj=A), 5)
        self.assertEqual(X('cls_method', 2, y=3).eval(obj=A), 5)
        self.assertEqual(X('cls_method', x=2, y=3).eval(obj=A), 5)

        class A:
            """Test class"""

            def __init__(self, x=2):
                self.x = x

            def obj_method(self, y):
                """Instance method"""
                return self.x + y

        self.assertEqual(X(A(2).obj_method, 3).eval(), 5)
        self.assertEqual(X(A(2).obj_method, y=3).eval(), 5)
        self.assertEqual(X(A(x=2).obj_method, 3).eval(), 5)
        self.assertEqual(X(A(x=2).obj_method, y=3).eval(), 5)

        self.assertEqual(X('obj_method', A(2), 3).eval(), 5)
        self.assertEqual(X('obj_method', A(2), y=3).eval(), 5)
        self.assertEqual(X('obj_method', A(x=2), 3).eval(), 5)
        self.assertEqual(X('obj_method', A(x=2), y=3).eval(), 5)

        self.assertEqual(X('obj_method', 3).eval(obj=A(2)), 5)
        self.assertEqual(X('obj_method', y=3).eval(obj=A(2)), 5)
        self.assertEqual(X('obj_method', 3).eval(obj=A(x=2)), 5)
        self.assertEqual(X('obj_method', y=3).eval(obj=A(x=2)), 5)

        class A:
            """Test class"""

            def __init__(self, x=2):
                self.x = x

            def __call__(self, y):
                return self.x + y

        self.assertEqual(X(A(2), 3).eval(), 5)
        self.assertEqual(X(A(2), y=3).eval(), 5)
        self.assertEqual(X(A(x=2), 3).eval(), 5)
        self.assertEqual(X(A(x=2), y=3).eval(), 5)

        # Setter
        x = X(function)
        try:
            # noinspection PyPropertyAccess
            x.callee = function
        except AttributeError:
            pass
        else:
            raise AttributeError

        x.callee_args = (2, 3, 4)
        self.assertTupleEqual(x.callee_args, (2, 3, 4))
        x.callee_kwargs = {'x': 2, 'y': 3, 'z': 4}
        self.assertDictEqual(x.callee_kwargs, {'x': 2, 'y': 3, 'z': 4})

    def test_Parser(self):
        """Unittest: Parser"""

        # A callable
        def function(x, y):
            """Test function"""
            return x + y

        # ---------------------------------------------------------------------
        # Unittest: Parser.tuple

        # Callee is mandatory
        try:
            Parser.tuple(tuple())
        except TypeError:
            pass
        else:
            raise AssertionError

        # No args, no kwargs
        self.assertTrue(cmp(Parser.tuple((function,)), (function, (), {})))
        self.assertTrue(cmp(Parser.tuple(('function',)), ('function', (), {})))
        try:
            Parser.tuple((123456,))
        except TypeError:
            pass
        else:
            # noinspection PyStatementEffect
            AssertionError

        # Only args
        self.assertTrue(cmp(Parser.tuple((function, (2, 3))),
                            (function, (2, 3), {})))
        self.assertTrue(cmp(Parser.tuple(('function', (2, 3))),
                            ('function', (2, 3), {})))

        # Only kwargs
        self.assertTrue(cmp(Parser.tuple((function, {'x': 2, 'y': 3})),
                            (function, (), {'x': 2, 'y': 3})))
        self.assertTrue(cmp(Parser.tuple(('function', {'x': 2, 'y': 3})),
                            ('function', (), {'x': 2, 'y': 3})))

        # args and kwargs
        self.assertTrue(cmp(Parser.tuple((function, (2,), {'y': 3})),
                            (function, (2,), {'y': 3})))
        self.assertTrue(cmp(Parser.tuple(('function', (2,), {'y': 3})),
                            ('function', (2,), {'y': 3})))

        # ---------------------------------------------------------------------
        # Unittest: Parser.list
        self.assertListEqual(Parser.list([F('a'), F('b'), F('c')]),
                             [F('a'), F('b'), F('c')])
        self.assertListEqual(Parser.list([function, function, function]),
                             [F(function), F(function), F(function)])
        self.assertListEqual(Parser.list(['a', 'b', 'c']),
                             [F('a'), F('b'), F('c')])
        self.assertListEqual(Parser.list([('a',), ('b',), ('c',)]),
                             [F('a'), F('b'), F('c')])

        self.assertListEqual(Parser.list([[['a'], 'b'], 'c']),
                             [F('a'), F('b'), F('c')])
        self.assertListEqual(Parser.list(['a', ['b', ['c']]]),
                             [F('a'), F('b'), F('c')])

        # ---------------------------------------------------------------------
        # Unittest: eval()
        self.assertListEqual(Parser.eval(F('a')), [F('a')])
        self.assertListEqual(Parser.eval(function), [F(function)])
        self.assertListEqual(Parser.eval('function'), [F('function')])
        self.assertListEqual(Parser.eval(('function',)), [F('function')])
        self.assertListEqual(Parser.eval(['function']), [F('function')])


# -----------------------------------------------------------------------------
# Execution
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding=UTF-8 -*-
# vim: fileencoding=UTF-8 tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -----------------------------------------------------------------------------
# Document Description
"""**Observer**

    The observer pattern is generally used to inform one or more registered
    objects (observers, subscribers, objects) about selected actions of an
    observed object (observable, publisher, subject).

    The time of activation is set to AFTER by default, i.e. the observable
    performs its own activity and then activates all registered observers in
    the specified order. This setting can be adjusted to before, after, both
    or even no activation at all via the parameter activate of Observable.

    This implementation provides several ways to decorate a function or class
    as an observable or observer.

        -   Observable Decoration
        -   Observer Decoration
        -   Class Decoration
        -   Instance Decoration

    Attributes
    ----------
    Observer (class):                           (uses BaseObserver)
        Creates an observer instance as a callable object.

    Observable (class):                         (uses BaseObservable)
        Creates an observable callable object exposing the pattern interface.

    BaseObserver (class):                       (for inheritance, only)
        A base implementation of the (abstract) observer base class.

    BaseObservable (class):                     (for inheritance, only)
        A base implementation of the (abstract) observable base class.

    Methods
    -------
        None.

    Examples
    --------

    A) Observable Decoration

    The simplest and at the same time the most pythonic variant of decoration
    is to decorate only the observed entities as a Observable.

    This is possible because all observer pattern functionalities are
    concentrated in the Observable.BaseClass = BaseObservable of the observable
    decorator, while the Observer.BaseClass = BaseObserver of the observer
    decorator remains empty here. If necessary, it is possible to inherit from
    both BaseClasses to modify their functionalities.

    ---------------------------------------------------------------------------
    from decoratory.observer import Observable
    from decoratory.basic import F

    def person(say: str = "Hello?"):
        print(f"{person.__name__} says '{say}'")

    @Observable(observers=F(person, 'Hey, dog!'))
    def dog(act: str = "Woof!"):
        print(f"{dog.__name__} acts '{act}'")

    # Case 1: Observable decoration
    #    ---> Person as an observer to observable dog
    person()                        # person says 'Hello?'    (person acting)
    dog()                           # dog acts 'Woof!'        (dog acting)
                                    # person says 'Hey, dog!' (observer to dog)

    def person(say: str = "Hello?"):
        print(f"{person.__name__} says '{say}'")

    @Observable(observers=F(person, 'Hey, cat!'))
    def cat(act: str = "Meow!"):
        print(f"{cat.__name__} acts '{act}'")

    @Observable(observers=[F(cat, 'Roar!'), F(person, 'Hey, dog!')])
    def dog(act: str = "Woof!"):
        print(f"{dog.__name__} acts '{act}'")

    # Case 2: Stacked observable decoration
    #    ---> Cat observes dog, person observes cat and dog
    person()                        # person says 'Hello?'    (person acting)

    cat()                           # cat acts 'Meow!'        (cat acting)
                                    # person says 'Hey, cat!' (observer to cat)

    dog()                           # dog acts 'Woof!'        (dog acting)
                                    # cat acts 'Roar!'        (observer to dog)
                                    # person says 'Hey, cat!' (observer to cat)
                                    # person says 'Hey, dog!' (observer to dog)

    @Observable(observers=[F(person, 'Hey, dog!'), F(cat, 'Roar!')])
    def dog(act: str = "Woof!"):
        print(f"{dog.__name__} acts '{act}'")

    # Case 3: Stacked observable decoration
    #    ---> Cat observes dog, person observes dog and cat
    dog()                           # dog acts 'Woof!'        (dog acting)
                                    # person says 'Hey, dog!' (observer to dog)
                                    # cat acts 'Roar!'        (observer to dog)
                                    # person says 'Hey, cat!' (observer to cat)

    @Observable(observers=[F(cat.substitute.callee, 'Roar!'),
                           F(person, 'Hey, dog!')])
    def dog(act: str = "Woof!"):
        print(f"{dog.__name__} acts '{act}'")

    # Case 4: Stacked observable decoration
    #    ---> Original cat observes dog, person observes dog and cat
    dog()                           # dog acts 'Woof!'        (dog acting)
                                    # cat acts 'Roar!'        (observer to dog)
                                    # person says 'Hey, dog!' (observer to dog)


    B) Observer Decoration

    In this reversed decoration scheme, the observer decorator collects its
    observables. This seems more elaborate at first glance, but some prefer to
    explicitly designate the Observer and Observable roles in their code.

    Because an observer decoration uses observable methods, all observable(s)
    must always be declared before their observer(s).

        1. Rule: Declare Observables before Observers

    For multiple decorations, the order of decoration is also relevant here.
    Each observable must be decorated before it is used by an observer.

        2. Rule: Decorating as @Observable before using in an @Observer

    ---------------------------------------------------------------------------
    from decoratory.observer import Observer, Observable
    from decoratory.basic import X

    @Observable
    def dog(act: str = "Woof!"):    # 1. Rule: declare dog before person!
        print(f"{dog.__name__} acts '{act}'")

    @Observer(observables=X(dog, 'Hey, dog!'))
    def person(say: str = "Hello?"):
        print(f"{person.__name__} says '{say}'")

    # Case 1: Observer decoration
    #    ---> Person as an observer to observable dog
    person()                        # person says 'Hello?'
    dog()                           # dog acts 'Woof!'        (dog acting)
                                    # person says 'Hey, dog!' (observer to dog)

    @Observable                     # 2. Rule: dog before cat & person
    def dog(act: str = "Woof!"):    # 1. Rule: dog before cat & person
        print(f"{dog.__name__} acts '{act}'")

    @Observer(observables=X(dog, 'Roar!'))
    @Observable                     # 2. Rule: observable cat before person
    def cat(act: str = "Meow!"):    # 1. Rule: cat before person
        print(f"{cat.__name__} acts '{act}'")

    @Observer(observables=[X(dog, 'Hey, dog!'),
                           X(cat.substitute.callee, say='Hey, cat!')])
    def person(say: str = "Hello?"):
        print(f"{person.__name__} says '{say}'")

    # Case 2: Stacked observer decoration
    #    ---> Cat observes dog, person observes cat and dog
    person()                        # person says 'Hello?'    (person acting)

    cat()                           # cat acts 'Meow!'        (cat acting)
                                    # person says 'Hey, cat!' (observer to cat)

    dog()                           # dog acts 'Woof!'        (dog acting)
                                    # cat acts 'Roar!'        (observer to dog)
                                    # person says 'Hey, cat!' (observer to cat)
                                    # person says 'Hey, dog!' (observer to dog)

    @Observable                     # 2. Rule: dog before cat & person
    def dog(act: str = "Woof!"):    # 1. Rule: dog before cat & person
        print(f"{dog.__name__} acts '{act}'")

    @Observable                     # 2. Rule: cat before person
    @Observer(observables=X(dog, 'Roar!'))
    def cat(act: str = "Meow!"):    # 1. Rule: cat before person
        print(f"{cat.__name__} acts '{act}'")

    @Observer(observables=[X(dog, 'Hey, dog!'), X(cat, say='Hey, cat!')])
    def person(say: str = "Hello?"):
        print(f"{person.__name__} says '{say}'")

    # Case 3: Stacked observer decoration
    #    ---> Cat observes dog, person observes cat and dog
    person()                        # person says 'Hello?'    (person acting)

    cat()                           # cat acts 'Meow!'        (cat acting)
                                    # person says 'Hey, cat!' (observer to cat)

    dog()                           # dog acts 'Woof!'        (dog acting)
                                    # cat acts 'Roar!'        (observer to dog)
                                    # person says 'Hey, dog!' (observer to dog)


    C) Static Class Decoration

    Both techniques, Observable Decoration and Observer Decoration, are static,
    in the sense, decorations are done e.g. in @-notation evaluated at compile
    time. They are applied to static functions.

    Decoration of a class by default addresses decoration of the class
    constructor, this means

    @Observable
    class Dog:
        def __init__(self):
            pass                    # Some code ...

    should be understood as

    class Dog:
        @Observable
        def __init__(self):
            pass                    # Some code ...

    ---------------------------------------------------------------------------
    from decoratory.observer import Observable

    class Person:
        def __init__(self, name: str = "Jane Doe"):
            print(f"{name} arrived.")

    @Observable(observers=Person)
    class Dog:
        def __init__(self, name: str = "Teddy"):
            print(f"Dog {name} arrived.")

    # Case 1: Dog is an observable to Person
    prs = Person()                  # Jane Doe arrived.
    dog = Dog()                     # Dog Teddy arrived.
                                    # Jane Doe arrived.

    class Person:
        def __init__(self, name: str = "Jane Doe"):
            print(f"{name} arrived.")

        @staticmethod
        def action1(act: str = "Hello?"):
            print(f"Person says {act}")

        @classmethod
        def action2(cls, act: str = "What's up?"):
            print(f"Person says {act}")

    @Observable(observers=[Person.action1, Person.action2])
    class Dog:
        def __init__(self, name: str = "Teddy"):
            print(f"Dog {name} arrived.")

    # Case 2: Dog is an observable to Person.action
    prs = Person()                  # Jane Doe arrived.
    dog = Dog()                     # Dog Teddy arrived.
                                    # Person says Hello?
                                    # Person says What's up?

    ---------------------------------------------------------------------------
    from decoratory.observer import Observable
    from decoratory.basic import F

    class Person:
        def __init__(self, name: str = "Jane Doe"):
            self.name = name
            print(f"{name} arrived.")

        def action(self, act: str = "Hello?"):
            print(f"{self.name} says {act}")

    prs1 = Person()                 # Jane Doe arrived.
    prs2 = Person("John Doe")       # John Doe arrived.

    @Observable(observers=[prs1.action, F(prs2.action, "What's up?")])
    class Dog:
        def __init__(self, name: str = "Teddy"):
            print(f"Dog {name} arrived.")

    # Case 3: Dog is an observable to actions of various person instances.
    dog = Dog()                     # Dog Teddy arrived.
                                    # Jane Doe says Hello?
                                    # John Doe says What's up?

    class Person:
        def __init__(self, name: str = "Jane Doe"):
            self.name = name
            print(f"{name} arrived.")

        @classmethod
        def actionA(cls, act: str = "Hello?"):
            print(f"Person says {act}")

        def actionB(self, act: str = "Hello?"):
            print(f"{self.name} says {act}")

    @Observable(methods=["action1", "action2"],
                observers=[Person.actionA, Person("Any Doe").actionB])
    class Dog:
        def __init__(self, name: str = "Teddy"):
            self.name = name
            print(f"Dog {name} arrived.")

        @staticmethod
        def action1(act: str = "Woof!"):
            print(f"Dog acts {act}")

        def action2(self, act: str = "Brrr!"):
            print(f"{self.name} acts {act}")

    # Case 4: Dog is an observable with selected actions.
                                    # Any Doe arrived.
    prs = Person()                  # Jane Doe arrived.
    dog = Dog()                     # Dog Teddy arrived.

    dog.action1()                   # Dog acts Woof!        (@staticmethod)
                                    # Person says Hello?    (@classmethod)
                                    # Any Doe says Hello?   (Instance 'Any')

    Dog.action2(dog)                # Teddy acts Brrr!      (Instance 'Teddy')
                                    # Person says Hello?    (@classmethod)
                                    # Any Doe says Hello?   (Instance 'Any')


    D) Dynamic Class Decoration

    The classic way to exchange information between objects with the observer
    pattern is through the active use of the register, dispatch, and unregister
    interface methods that an observable exposes. Information can be given to
    the right recipients at relevant places in the code. For this, the classes
    are not decorated and dynamic decoration comes into play. Dynamic
    decoration is used, often also in connection with getter/setter/property
    constructions, since data changes take place meaningfully over these
    methods.

    class Note:                             # Observer without decoration!
        def info(self, thing):
            print(f"Note.info: val = {thing.a}")

    class Thing:                            # Observable without decoration!
        def __init__(self, a=0):
            self._a = a
        def inc(self):
            self._a += 1
        def get_a(self):
            return self._a
        def set_a(self, value):
            self._a = value
        a = property(get_a, set_a)

    ---------------------------------------------------------------------------
    from decoratory.observer import Observable
    from decoratory.basic import F

    # (1) Setup instances
    nti = Note()                    # Note instance
    tgi = Thing()                   # Thing instance

    # (2) Dynamic decoration of some methods: Late binding
    tgi.inc = Observable(tgi.inc)           # Late method decoration
    Thing.set_a = Observable(Thing.set_a)   # Late property decoration
    Thing.a = property(Thing.get_a, Thing.set_a)

    # (3) Register the observer (Note) with the observable (Thing)
    tgi.inc.observable.register(F(nti.info, tgi))
    tgi.set_a.observable.register(F(nti.info, thing=tgi))

    # Case 1: Change self.a = 0 using inc()
    tgi.inc()                       # Note.info: val = 1

    # Case 2: Change self.a = 1 using setter via property
    tgi.a = 2                       # Note.info: val = 2

    # Case 3: Notification from inc() to nti.info() about Thing(3)
    tgi.inc.observable.dispatch(nti.info, Thing(3))
                                    # Note.info: val = 3

    # Case 4: Notification from set_a() to nti.info() about Thing(4)
    tgi.set_a.observable.dispatch(nti.info, Thing(4))
                                    # Note.info: val = 4

    # Case 5: Print the current value of tgi.a
    print(f"a = {tgi.a}")           # a = 2     (no changes by notification)

    # Case 6: Print list of all observers
    print(tgi.inc.observable.observers(classbased=True))
    # ---> {'Note': ['F(info, <__main__.Thing object at ..)']}
    print(tgi.set_a.observable.observers(classbased=True))
    # ---> {'Note': ['F(info, thing=<__main__.Thing object at ..)']}

    # Case 7: Unregister nti.info from tgi
    tgi.inc.observable.unregister(nti.info)
    print(tgi.inc.observable.observers(classbased=True))    # {}
"""

# -----------------------------------------------------------------------------
# Module Level Dunders
__title__ = "Observer"
__module__ = "observer.py"
__author__ = "Martin Abel"
__maintainer__ = "Martin Abel"
__credits__ = ["Martin Abel"]
__company__ = "eVation"
__email__ = "python@evation.eu"
__url__ = "http://evation.eu"
__copyright__ = f"(c) copyright 2020-2023, {__company__}"
__created__ = "2020-01-01"
__version__ = "0.9.0.2"
__date__ = "2023-06-21"
__time__ = "19:34:05"
__state__ = "Beta"
__license__ = "PSF"

__all__ = ["Observer", "BaseObserver", "Observable", "BaseObservable"]

# -----------------------------------------------------------------------------
# Libraries & Modules
from functools import update_wrapper
from typing import Union
from decoratory.basic import Activation, F, Parser, X


# -----------------------------------------------------------------------------
# Classes
class BaseObservable:
    """**BaseObservable**

    A base implementation of the (abstract) observable base class. It manages
    (abstract) observers (of F-type) within a private dictionary using the
    methods:

    register  : Register an observer for callback
    unregister: Unregister an observer
    dispatch  : Dispatch a given observer or even all observers
    observers : Dictionary of all registered observers (of F-type)

    While the methods register, unregister and observers only handles given
    data into F objects and/or collects them, the dispatch method applies F's
    eval method (without arguments!) to them. It's in the user's responsibility
    to make sure that these calls succeed, i.e. for a class or instance/object
    method of class A or an instance a = A():
     - Registration call:   F(a.method, *args, **kwargs).eval()         or
                            F('method', a, *args, **kwargs).eval()      but not
     - Dynamic call:        F('method', *args, **kwargs).eval(obj=a)
    """

    def __init__(self, *args, **kwargs) -> None:
        self.args = args or tuple()
        self.kwargs = kwargs or dict()

        self.__observers = dict()  # dict of F-type observers: callee is key!

    # Methods of the Observer Pattern
    def register(self,
                 observer: Union[F, callable, str],
                 *observer_args: object,
                 **observer_kwargs: object) -> None:
        """Register a function (callable) or method (str) for callback.

        Parameters:
            observer (F|callable|str): Callback function|method of the observer
            observer_args (object): Callback (default) positional arguments
            observer_kwargs (object): Callback (default) keyword arguments

        Returns:
            None.
        """
        if isinstance(observer, F):
            if observer_args or observer_kwargs:
                observer.callee_args = observer_args
                observer.callee_kwargs = observer_kwargs
            self.__observers[observer.callee] = observer  # Override mode
        elif callable(observer) or isinstance(observer, str):
            obs = F(observer, *observer_args, **observer_kwargs)
            self.__observers[observer] = obs
        else:
            raise TypeError(f"'{observer}' cannot be registered.")

    def unregister(self,
                   observer: Union[F, callable, str, None] = None) -> None:
        """Unregister an observer.

        If the observer parameter is omitted (None), all registered observers
        will be unregistered.

        Parameters:
            observer (F|callable|str|None): Callback to be unregistered

        Returns:
            None.
        """
        if observer is None:
            self.__observers.clear()
        elif isinstance(observer, F):
            self.__observers.pop(observer.callee, None)  # Quiet mode
        elif callable(observer) or isinstance(observer, str):
            self.__observers.pop(observer, None)  # Quiet mode
        else:
            raise TypeError(f"'{observer}' cannot be unregistered.")

    def dispatch(self,
                 observer: Union[F, callable, str, None] = None,
                 *observer_args: object,
                 **observer_kwargs: object) -> None:
        """Dispatch an observer.

        If the observer parameter is omitted (None), all registered observers
        will be dispatched.

        Parameters:
            observer (F|callable|str|None): Callback to be dispatched
            observer_args (object): Callback (default) positional arguments
            observer_kwargs (object): Callback (default) keyword arguments

        Returns:
            None.
        """
        if observer is None:
            # Registration call using default arguments, no extra eval obj!
            if observer_args or observer_kwargs:
                for obs in self.__observers.values():
                    F(obs.callee, *observer_args, **observer_kwargs).eval()
            else:
                for obs in self.__observers.values():
                    obs.eval()
        elif isinstance(observer, F):
            # Dynamic call using current/default arguments, no extra eval obj!
            if observer_args or observer_kwargs:
                F(observer.callee, *observer_args, **observer_kwargs).eval()
            else:
                observer.eval()
        elif callable(observer):
            # Dynamic call using current arguments, no extra eval obj!
            F(observer, *observer_args, **observer_kwargs).eval()
        elif observer in self.__observers:
            # Try to resolve the string or something else...
            observer = self.__observers.get(observer)
            if observer_args or observer_kwargs:
                F(observer.callee, *observer_args, **observer_kwargs).eval()
            else:
                observer.eval()
        else:
            raise TypeError(f"'{observer}' cannot be dispatched.")

    def observers(self, classbased: bool = False) -> dict:
        """Listing of all observers.

        Observers are collected in a dict, which is returned by default with
        classbased=False. Calling with classbased=True returns a dictionary
        with key-value-pair syntax {classname: list(methods)}.

        Parameters:
            classbased (bool): A boolean switch for returned data structure

        Returns:
            observers (dict): Dictionary of all observers
        """
        if bool(classbased):
            result = dict()
            for obs in self.__observers.values():
                *skip, cls, mtd = obs.callee.__qualname__.split(".")
                result.setdefault(cls, []).append(repr(obs))
            return result
        else:
            return self.__observers  # Has to be the default (without params)!


class Observable:
    """**Observable** (Publisher, Subject)

    Creating an observable instantiates a callable object which exposes the
    four basic observable pattern methods register, unregister, dispatch and
    observers via an observable attribute for an instance of BaseClass
    (default = BaseObservable) as well as the original decorator arguments,
    if present, like the callable to be substituted, observers, methods and
    activation point in time.

    Observable(substitute, *args, observers, methods, activate, **kwargs)

    Attributes
    ----------
    substitute (callable|type|None):
        A type to be made an observable

    observers (list|F|callable|str|None):
        (List of) callable(s) of observers

    methods (list|F|callable|str|None):
        (List of) callable(s) of as method name strings

    activate (Activation):
        Dispatch activation point in time

    Methods
    -------
        None.
    """

    BaseClass = BaseObservable

    def __init__(self,
                 substitute: Union[type, callable, None] = None,
                 *args: object,
                 observers: Union[list, callable, F, str, None] = None,
                 methods: Union[list, callable, F, str, None] = None,
                 activate: Activation = Activation.AFTER,
                 **kwargs: object) -> None:
        """Observable (Publisher, Subject).

        Parameters:
            substitute (callable|type): Callable|Type to be made an observable
            observers (list|F|callable|str): (List of) callable(s) of observers
            methods (list|F|callable|str): (List of) callable(s) of as strings
            activate (Activation): Dispatch activation point in time

        Returns:
            None.
        """
        self.__set__substitute(substitute)
        self.__set__observers(observers)
        self.__set__methods(methods)
        self.__set__activate(activate)

        # --- Decorator Arguments Pattern (1/2)
        if self.__get__substitute() is not None:
            # Decoration without parameter(s)
            self.__set__substitute(
                F(self.__get__substitute(), *args, **kwargs))
            update_wrapper(self, self.__get__substitute().callee, updated=())

            self.__set__observable(Observable.BaseClass())
        else:
            # Decoration with parameter(s)
            self.__set__observable(Observable.BaseClass(*args, **kwargs))

    def __call__(self, *args, **kwargs):
        # --- Decorator Arguments Pattern (2/2)
        if self.__get__substitute() is None:
            # Decoration with parameter(s)
            self.__set__substitute(F(args[0], *args[1:], **kwargs))

            # Decoration of a type means decoration of *all* submitted methods
            if self.__get__methods():
                # Resolve list of methods:
                subst = self.__get__substitute().callee
                for mtd, mtd_args, mtd_kwargs in self.__get__methods():
                    if isinstance(mtd, str) and hasattr(subst, mtd):
                        mtds = mtd
                        mtd0 = getattr(subst, mtds)
                    elif callable(mtd):
                        mtds = mtd.__name__
                        mtd0 = mtd
                    else:
                        raise TypeError(f"{mtd} is nor a string nor callable.")
                    # noinspection PyArgumentEqualDefault
                    mtd1 = Observable(
                        None,  # Call with arguments (substitute is None)
                        *self.observable.args,
                        observers=self.__get__observers(),
                        methods=None,  # Resolved, call to else part below!
                        activate=self.__get__activate(),
                        **self.observable.kwargs)(
                        mtd0, *mtd_args, **mtd_kwargs)
                    setattr(subst, mtds, mtd1)

                # Return the undecorated original class
                return subst
            else:
                # Setup observers
                if self.__get__observers():
                    for observer in self.__get__observers():
                        self.__get__observable().register(observer)

                # Complete wrapper and return observable
                update_wrapper(self, self.__get__substitute().callee,
                               updated=())
                return self
        else:  # *** Wrapper ***
            # Dispatch BEFORE
            if self.__get__activate() & Activation.BEFORE:
                self.__get__observable().dispatch()

            # Delegation: apply the substitute, current before default values
            if args or kwargs:
                result = F(self.__get__substitute().callee, *args,
                           **kwargs).eval()
            else:
                result = self.__get__substitute().eval()

            # Dispatch AFTER
            if self.__get__activate() & Activation.AFTER:
                self.__get__observable().dispatch()

            return result

    # Getter, Setter, Properties
    def __get__substitute(self):
        return self.__substitute

    def __set__substitute(self, value):
        self.__substitute = value

    substitute = property(__get__substitute)

    def __get__observers(self):
        return self.__observers

    def __set__observers(self, value):
        self.__observers: list = Parser.eval(value)

    observers = property(__get__observers)

    def __get__methods(self):
        return self.__methods

    def __set__methods(self, value):
        self.__methods: list = Parser.eval(value)

    methods = property(__get__methods)

    def __get__observable(self):
        return self.__observable

    def __set__observable(self, value):
        self.__observable = value

    observable = property(__get__observable)

    def __get__activate(self):
        return self.__activate

    def __set__activate(self, activate):
        self.__activate = activate if isinstance(
            activate, Activation) else Activation.NONE

    activate = property(__get__activate, __set__activate)


class BaseObserver:
    """**BaseObserver**

    A base implementation of the (abstract) observer base class.

    As long as this class, just like here, is an empty dummy, decoration of a
    callable as an observer is optional. If BaseObserver is overwritten and
    assigned to the observers BaseClass attribute all non captured decorator
    args & kwargs will be submitted to be used in customized class
    functionalities.
    """

    def __init__(self, *args, **kwargs):
        self.args = args or tuple()
        self.kwargs = kwargs or dict()


class Observer:
    """**Observer** (Subscriber, Object)

    Creating an observer instantiates a callable object which exposes the
    original decorator arguments, if present, like the callable to be
    substituted, observables and methods.

    Observer(substitute, *args, observables, methods, **kwargs)

    Attributes
    ----------
    substitute (callable|type|None):
        A type to be made an observer

    observables (list|X|callable|str|None):
        (List of) callable(s) of observables

    methods (list|X|callable|str|None):
        (List of) callable(s) of as method name strings

    Methods
    -------
        None.
   """

    BaseClass = BaseObserver

    def __init__(self,
                 substitute: Union[type, callable, None] = None,
                 *args: object,
                 observables: Union[list, callable, X, str, None] = None,
                 methods: Union[list, callable, X, str, None] = None,
                 **kwargs: object) -> None:
        """**Observer** (Subscriber, Object)

        Parameters:
            substitute (callable|type): Callable|Type to be made an observable
            observables (list|X|callable|str): (List of) callable(s) of observables
            methods (list|X|callable|str): (List of) callable(s) of as strings

        Returns:
            None.
        """
        self.__set__substitute(substitute)
        self.__set__observables(observables)
        self.__set__methods(methods)

        # --- Decorator Arguments Pattern (1/2)
        if self.__get__substitute() is not None:
            # Decoration without parameter(s)
            self.__set__substitute(
                F(self.__get__substitute(), *args, **kwargs))
            update_wrapper(self, self.__get__substitute().callee, updated=())

            self.__set__observer(Observer.BaseClass())
        else:
            # Decoration with parameter(s)
            self.__set__observer(Observer.BaseClass(*args, **kwargs))

    def __call__(self, *args, **kwargs):
        # --- Decorator Arguments Pattern (2/2)
        if self.__get__substitute() is None:
            # Decoration with parameter(s)
            self.__set__substitute(F(args[0], *args[1:], **kwargs))

            # Decoration of a type means decoration of *all* submitted methods
            if self.__get__methods():
                # Resolve list of methods:
                subst = self.__get__substitute().callee
                for mtd, mtd_args, mtd_kwargs in self.__get__methods():
                    if isinstance(mtd, str) and hasattr(subst, mtd):
                        mtds = mtd
                        mtd0 = getattr(subst, mtds)
                    elif callable(mtd):
                        mtds = mtd.__name__
                        mtd0 = mtd
                    else:
                        raise TypeError(f"{mtd} is nor a string nor callable.")
                    # noinspection PyArgumentEqualDefault
                    mtd1 = Observer(
                        None,  # Call with deco arguments (substitute is None)
                        *self.observer.args,
                        observables=self.__get__observables(),
                        methods=None,  # Resolved, call to else part below!
                        **self.observer.kwargs)(
                        mtd0, *mtd_args, **mtd_kwargs)
                    setattr(subst, mtds, mtd1)

                # Return the undecorated original class
                return subst
            else:
                # Register self as a callable object for callback
                # CAUTION: observables is a list of X-objects with semantics
                #     obs = X(observABLE, observER_args, observER_kwargs)
                # The arguments belong to the observer (self) but not to the
                # observable from the observables list!
                if self.__get__observables():
                    for observable in self.__get__observables():
                        if isinstance(observable.callee, Observable):
                            observable.callee.observable.register(
                                self, *observable.callee_args,
                                **observable.callee_kwargs)
                        else:
                            raise TypeError(
                                f"{observable.callee} is not an observable.")

                # Complete wrapper and return observer
                update_wrapper(self, self.__get__substitute().callee,
                               updated=())
                return self
        else:  # *** Wrapper ***
            # Delegation: apply the substitute, current before default values
            if args or kwargs:
                return F(self.__get__substitute().callee, *args,
                         **kwargs).eval()
            else:
                return self.__get__substitute().eval()

    # Getter, Setter, Properties
    def __get__substitute(self):
        return self.__substitute

    def __set__substitute(self, value):
        self.__substitute = value

    substitute = property(__get__substitute)

    def __get__observables(self):
        return self.__observables

    def __set__observables(self, value):
        self.__observables: list = Parser.eval(value)

    observables = property(__get__observables)

    def __get__methods(self):
        return self.__methods

    def __set__methods(self, value):
        self.__methods: list = Parser.eval(value)

    methods = property(__get__methods)

    def __get__observer(self):
        return self.__observer

    def __set__observer(self, value):
        self.__observer = value

    observer = property(__get__observer)


# -----------------------------------------------------------------------------
# Simple example
if __name__ == '__main__':
    from decoratory.banner import __banner as banner
    import decoratory.observer as module

    banner(title=__title__,
           version=__version__,
           date=__date__,
           time=__time__,
           docs=(module, Observer, Observable, BaseObserver, BaseObservable),
           author=__author__,
           maintainer=__maintainer__,
           company=__company__,
           email=__email__,
           url=__url__,
           copyright=__copyright__,
           state=__state__,
           license=__license__)



#!/usr/bin/env python
# -*- coding=UTF-8 -*-
# vim: fileencoding=UTF-8 tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -----------------------------------------------------------------------------
# Document Description
"""**Wrapper**

    A wrapper encloses the original function with an

     - (optional) before call functionality, and/or
     - (optional) after  call functionality, and/or

    substitutes the original functionality with an

     - (optional) replace call functionality.

    All three decorator parameters, before, after and replace, can be combined
    with each other and support both single callables and (nested) lists of
    F-types (imported from module decoratory.basic, see below for details).
    In addition, replace supports passing results from successive replacement
    calls through an optional keyword argument named result (default value is
    None).

    Attributes
    ----------
    Wrapper (class):
        Creates a wrapper instance as a callable object.

    Methods
    -------
        None.

    Examples
    --------

    # -------------------------------------------------------------------------
    from decoratory.wrapper import Wrapper

    # Case 1: Dynamic decoration with decorator arguments, only
    def some_function(value: str = "original"):
        print(f"value = '{value}'")

    # Function call with default parameters
    some_function()                 # value = 'original'
    some_function = Wrapper(some_function, value="changed")
    some_function()                 # value = 'changed'

    # -------------------------------------------------------------------------
    from decoratory.wrapper import Wrapper
    from decoratory.basic import F

    # Case 2: Decoration with before and after functionalities
    def print_message(message: str = "ENTER"):
        print(message)

    @Wrapper(before=print_message, after=F(print_message, "LEAVE"))
    def some_function(value: str = "original"):
        print(f"value = '{value}'")

    some_function()                 # ENTER
                                    # value = 'original'
                                    # LEAVE

    # Case 3: Decoration with replace functionality
    def replace_wrapper(value: str="replace"):
        # 1. Edit the call parameters for the original functionality
        value = value.upper()
        # 2. Execute original functionality with modified call parameters
        result = some_function.substitute.callee(value)             # (1)
        # 3. Edit the result and return this modified result
        return f"result: '{result}'"

    @Wrapper(replace=replace_wrapper)
    def some_function(value: str = "original"):
        print(f"value = '{value}'")
        return value

    result = some_function()        # value = 'REPLACE'
    print(result)                   # result: 'REPLACE'

    # Case 4: Decoration with before, after and multiple replacements
    def print_message(message: str = "UNDEFINED"):
        print(message)

    def replacement_printer(add: int = 1, *, result=None):
        result += add if isinstance(result, int) else 0
        print(f"result = {result}")
        return result

    @Wrapper(before=F(print, "ENTER"), # Python's print()
             replace=[F(replacement_printer, 1, result=0),
                      F(replacement_printer, 3),
                      F(replacement_printer, 5)],
             after=F(print_message, "LEAVE"))
    def result_printer(message: str = "UNKNOWN"):
        print(message)

    result_printer()                # ENTER         (before)
                                    # result = 1    (replacement_printer, 1)
                                    # result = 4    (replacement_printer, 3)
                                    # result = 9    (replacement_printer, 5)
                                    # LEAVE         (after)
                                    # 9             (output default_printer)

    @Wrapper(before=F(print, "BEFORE init"), after=F(print, "AFTER init"))
    class Animal:
        def __init__(self, name):
            self.name = name
            print("RUNNING init")

    # Case 5: Decoration of a class always refers to __init__
    a = Animal(name='Teddy')        # BEFORE init
                                    # RUNNING init
                                    # AFTER init

    # Case 6: Define a private wrapper library
    before_wrapper = Wrapper(before=F(print, "BEFORE"))
    after_wrapper = Wrapper(after=F(print, "AFTER"))

    # Multiple decorations for specialized functionality encapsulation
    @before_wrapper
    @after_wrapper
    def some_function(value: str = "original"):
        print(f"value = '{value}'")

    some_function()                 # BEFORE
                                    # value = 'original'
                                    # AFTER
"""

# -----------------------------------------------------------------------------
# Module Level Dunders
__title__ = "Wrapper"
__module__ = "wrapper.py"
__author__ = "Martin Abel"
__maintainer__ = "Martin Abel"
__credits__ = ["Martin Abel"]
__company__ = "eVation"
__email__ = "python@evation.eu"
__url__ = "http://evation.eu"
__copyright__ = f"(c) copyright 2020-2023, {__company__}"
__created__ = "2020-01-01"
__version__ = "0.9.0.2"
__date__ = "2023-06-21"
__time__ = "12:48:50"
__state__ = "Beta"
__license__ = "PSF"

__all__ = ["Wrapper"]

# -----------------------------------------------------------------------------
# Libraries & Modules
from functools import update_wrapper
from typing import Union
from decoratory.basic import F, Parser


# -----------------------------------------------------------------------------
# Classes
class Wrapper:
    """**Wrapper**

    Wrapper(substitute, *args, before, replace, after, **kwargs)

    Attributes
    ----------
    substitute (callable|type):
        A type to be made a wrapper

    before (callable|list):
        (List of) callable(s) to be executed before substitute

    replace (callable|list):
        (List of) callable(s) replacing the substitute

    after (callable|list):
        (List of) callable(s) to be executed after substitute

    Methods
    -------
        None.
    """

    def __init__(self,
                 substitute: Union[type, callable, None] = None,
                 *args: object,
                 before: Union[callable, list, None] = None,
                 replace: Union[callable, list, None] = None,
                 after: Union[callable, list, None] = None,
                 **kwargs: object) -> None:
        """Set up a wrapper.

        Parameters:
            substitute (object): A type to be made a wrapper.
            before (object): (List of) callable(s) to be executed before
            replace (object): (List of) callable(s) replacing the substitute
            after (object): (List of) callable(s) to be executed after

        Returns:
            self (object): Wrapper decorator instance
        """
        self.__set__substitute(substitute)
        self.__set__before(before)
        self.__set__replace(replace)
        self.__set__after(after)

        # --- Decorator Arguments Pattern (1/2)
        if self.__get__substitute() is not None:
            # Decoration without parameter(s)
            self.__set__substitute(
                F(self.__get__substitute(), *args, **kwargs))
            update_wrapper(self, self.__get__substitute().callee, updated=())

    def __call__(self, *args, **kwargs):
        # --- Decorator Arguments Pattern (2/2)
        if self.__get__substitute() is None:
            # Decoration with parameter(s)
            self.__set__substitute(F(args[0], *args[1:], **kwargs))
            update_wrapper(self, self.__get__substitute().callee, updated=())
            return self
        else:  # *** Wrapper ***
            # Action BEFORE
            if self.__get__before():
                for before in self.__get__before():
                    before.eval()

            # Delegation vs. REPLACE
            result = None
            if self.__get__replace():
                # Replacement
                if args or kwargs:
                    for replace in self.__get__replace():
                        d = replace.callee.__kwdefaults__
                        if d and 'result' in d:
                            d['result'] = result
                        result = F(replace.callee, *args, **kwargs).eval()
                else:
                    for replace in self.__get__replace():
                        d = replace.callee.__kwdefaults__
                        if d and 'result' in d:
                            d['result'] = result
                        result = replace.eval()
            else:
                # Delegation
                if args or kwargs:
                    result = F(self.__get__substitute().callee, *args,
                               **kwargs).eval()
                else:
                    result = self.__get__substitute().eval()

            # Action AFTER
            if self.__get__after():
                for after in self.__get__after():
                    after.eval()

            return result

    # Getter, Setter, Properties
    def __get__substitute(self):
        return self.__substitute

    def __set__substitute(self, value):
        self.__substitute = value

    substitute = property(__get__substitute)

    def __get__before(self):
        return self.__before

    def __set__before(self, value):
        self.__before: list = Parser.eval(value)

    before = property(__get__before)

    def __get__replace(self):
        return self.__replace

    def __set__replace(self, value):
        self.__replace: list = Parser.eval(value)

    replace = property(__get__replace)

    def __get__after(self):
        return self.__after

    def __set__after(self, value):
        self.__after: list = Parser.eval(value)

    after = property(__get__after)


# -----------------------------------------------------------------------------
# Simple example
if __name__ == '__main__':
    import decoratory.wrapper as module
    from decoratory.banner import __banner as banner

    banner(title=__title__,
           version=__version__,
           date=__date__,
           time=__time__,
           docs=(module, Wrapper),
           author=__author__,
           maintainer=__maintainer__,
           company=__company__,
           email=__email__,
           url=__url__,
           copyright=__copyright__,
           state=__state__,
           license=__license__)



#!/usr/bin/env python
# -*- coding=UTF-8 -*-
# vim: fileencoding=UTF-8 tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -----------------------------------------------------------------------------
# Document Description
"""**Test Singleton**"""

# -----------------------------------------------------------------------------
# Module Level Dunders
__title__ = "Test Singleton"
__module__ = "test_singleton.py"
__author__ = "Martin Abel"
__maintainer__ = "Martin Abel"
__credits__ = ["Martin Abel"]
__company__ = "eVation"
__email__ = "python@evation.eu"
__url__ = "http://evation.eu"
__copyright__ = f"(c) copyright 2020-2023, {__company__}"
__created__ = "2020-01-01"
__version__ = "0.9.0.1"
__date__ = "2023-06-21"
__time__ = "13:37:48"
__state__ = "Beta"
__license__ = "PSF"

# -----------------------------------------------------------------------------
# Libraries & Modules
import unittest

from decoratory.singleton import Singleton, SemiSingleton


# -----------------------------------------------------------------------------
# Test Class
# noinspection PyPep8Naming
class TestSingleton(unittest.TestCase):
    class TestClass:
        """A simple test class with some parameters"""

        def __init__(self, name: str, foo=None, bar="any stuff"):
            self.name = name
            self.foo = foo
            self.bar = bar

        def __repr__(self):
            return f"{self.__class__.__name__}(" \
                   f"'{self.name}', {self.foo}, '{self.bar}')"

    class AltClass:
        """An alternative test class with some parameters"""

        def __init__(self, name: str, foo=None, bar="other stuff"):
            self.name = name
            self.foo = foo
            self.bar = bar

        def __repr__(self):
            return f"{self.__class__.__name__}(" \
                   f"'{self.name}', {self.foo}, '{self.bar}')"

    def setUp(self):
        """Setup"""
        pass

    def tearDown(self):
        """Shutdown"""
        pass

    def test_decoration_without_brackets(self):
        """Unittest: Decoration without brackets"""

        TestClass = TestSingleton.TestClass
        TestClassSingleton = Singleton(TestClass)

        # Instances
        a = TestClassSingleton(name="a")  # New single object a
        b = TestClassSingleton(name="b")  # Singleton: b is a

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIs(a, b)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('a', None, 'any stuff')")
        self.assertEqual(a, b)
        self.assertEqual(TestClass.__doc__, TestClassSingleton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)

    def test_decoration_with_empty_brackets(self):
        """Unittest: Decoration with empty brackets"""

        TestClass = TestSingleton.TestClass
        TestClassSingleton = Singleton()(TestClass)

        # Instances
        a = TestClassSingleton(name="a")  # New single object a
        b = TestClassSingleton(name="b")  # Singleton: b is a

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIs(a, b)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('a', None, 'any stuff')")
        self.assertEqual(a, b)
        self.assertEqual(TestClass.__doc__, TestClassSingleton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)

    def test_decoration_with_default(self):
        """Unittest: Decoration with default"""

        TestClass = TestSingleton.TestClass
        TestClassSingleton = Singleton(TestClass, name="a")

        # Instances
        a = TestClassSingleton()  # New single object a (default)
        b = TestClassSingleton(name="b")  # Singleton: b is a

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIs(a, b)

        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('a', None, 'any stuff')")
        self.assertEqual(a, b)
        self.assertEqual(TestClass.__doc__, TestClassSingleton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)

    def test_decoration_with_resettable(self):
        """Unittest: Decoration with resettable"""

        # ---------------------------------------------------------------------
        # Default: resettable = False
        TestClass = TestSingleton.TestClass
        TestClassSingleton = Singleton(TestClass)

        # Instances
        a = TestClassSingleton(name="a")  # New single object a (default)

        # Tests
        self.assertFalse(hasattr(TestClass, "reset"))
        self.assertFalse(hasattr(TestClassSingleton, "reset"))
        self.assertFalse(hasattr(a, "reset"))

        # Test exception
        try:
            TestClassSingleton.reset()
        except AttributeError:
            pass
        else:
            raise AssertionError

        # ---------------------------------------------------------------------
        # Set: resettable = False
        TestClass = TestSingleton.TestClass
        # noinspection PyArgumentEqualDefault
        TestClassSingleton = Singleton(TestClass, resettable=False)

        # Instances
        a = TestClassSingleton(name="a")  # New single object a (default)

        # Tests
        self.assertFalse(hasattr(TestClass, "reset"))
        self.assertFalse(hasattr(TestClassSingleton, "reset"))
        self.assertFalse(hasattr(a, "reset"))

        # Test exception
        try:
            TestClassSingleton.reset()
        except AttributeError:
            pass
        else:
            raise AssertionError

        # ---------------------------------------------------------------------
        # Set: resettable = True
        TestClass = TestSingleton.TestClass
        TestClassSingleton = Singleton(TestClass, resettable=True)

        # Instances
        a = TestClassSingleton(name="a")  # New single object a (default)

        # Tests
        self.assertFalse(hasattr(TestClass, "reset"))
        self.assertTrue(hasattr(TestClassSingleton, "reset"))  # The only one!
        self.assertFalse(hasattr(a, "reset"))

        b = TestClassSingleton(name="b")  # Singleton: b is a
        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('a', None, 'any stuff')")
        self.assertEqual(a, b)
        self.assertIs(a, b)

        TestClassSingleton.reset()

        c = TestClassSingleton(name="c")  # New single object c (default)
        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('c', None, 'any stuff')")
        self.assertNotEqual(a, c)
        self.assertIsNot(a, c)

    def test_semi_singleton(self):
        """Unittest: Decoration as a semi singleton"""

        TestClass = TestSingleton.TestClass
        TestClassSingleton = SemiSingleton(TestClass)

        # Instances
        a = TestClassSingleton(name="a")  # New single object a (default)

        # Tests
        self.assertFalse(hasattr(TestClass, "reset"))
        self.assertTrue(hasattr(TestClassSingleton, "reset"))  # The only one!
        self.assertFalse(hasattr(a, "reset"))

        b = TestClassSingleton(name="b")  # Singleton: b is a
        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(b), "TestClass('a', None, 'any stuff')")
        self.assertEqual(a, b)
        self.assertIs(a, b)

        TestClassSingleton.reset()

        c = TestClassSingleton(name="c")  # New single object c (default)
        self.assertEqual(repr(a), "TestClass('a', None, 'any stuff')")
        self.assertEqual(repr(c), "TestClass('c', None, 'any stuff')")
        self.assertNotEqual(a, c)
        self.assertIsNot(a, c)

    def test_get_instance(self):
        """Unittest: Decoration get_instance"""

        TestClass = TestSingleton.TestClass
        TestClassSingleton = Singleton(TestClass)

        # Instance
        a = TestClassSingleton(name="a")  # New single object a (default)

        # Tests
        self.assertFalse(hasattr(TestClass, "get_instance"))
        self.assertTrue(hasattr(TestClassSingleton, "get_instance"))  # This!
        self.assertFalse(hasattr(a, "get_instance"))

        self.assertIs(TestClassSingleton.get_instance(), a)
        self.assertEqual(TestClassSingleton.get_instance(), a)

    def test_concurrency(self):
        """Unittest: Decoration concurrency
            --> Two parallel singletons do not interfere with each other!
        """

        TestClass = TestSingleton.TestClass
        TestClassSingleton = Singleton(TestClass)
        AltClass = TestSingleton.AltClass
        AltClassSingleton = Singleton()(AltClass)

        # Instances
        a = TestClassSingleton(name="a")  # New single object a
        b = TestClassSingleton(name="b")  # Singleton: b is a
        c = AltClassSingleton(name="c")  # New single object c
        d = AltClassSingleton(name="d")  # Singleton: d is c

        # Tests
        self.assertIsInstance(a, TestClass)
        self.assertIsInstance(b, TestClass)
        self.assertIsInstance(c, AltClass)
        self.assertIsInstance(d, AltClass)
        self.assertIs(a, b)
        self.assertIs(c, d)
        self.assertIsNot(a, c)
        self.assertIsNot(a, d)
        self.assertIsNot(b, c)
        self.assertIsNot(b, d)

        self.assertEqual(a, b)
        self.assertEqual(c, d)
        self.assertNotEqual(a, c)
        self.assertNotEqual(a, d)
        self.assertNotEqual(b, c)
        self.assertNotEqual(b, d)

        self.assertEqual(TestClass.__doc__, TestClassSingleton.__doc__)
        self.assertEqual(TestClass.__doc__, a.__doc__)
        self.assertEqual(TestClass.__doc__, b.__doc__)
        self.assertEqual(AltClass.__doc__, AltClassSingleton.__doc__)
        self.assertEqual(AltClass.__doc__, c.__doc__)
        self.assertEqual(AltClass.__doc__, d.__doc__)


# -----------------------------------------------------------------------------
# Execution
if __name__ == "__main__":
    unittest.main()
